r"""
Algebra of Scalar Fields

The class :class:`ScalarFieldAlgebra` implements the commutative algebra
`C^0(M)` of scalar fields on a topological manifold `M` over a topological
field `K`. By *scalar field*, it
is meant a continuous function `M \to K`. The set
`C^0(M)` is an algebra over `K`, whose ring product is the pointwise
multiplication of `K`-valued functions, which is clearly commutative.

AUTHORS:

- Eric Gourgoulhon, Michal Bejger (2014-2015): initial version
- Travis Scrimshaw (2016): review tweaks

REFERENCES:

- [Lee2011]_
- [KN1963]_
"""

#******************************************************************************
#       Copyright (C) 2015 Eric Gourgoulhon <eric.gourgoulhon@obspm.fr>
#       Copyright (C) 2015 Michal Bejger <bejger@camk.edu.pl>
#       Copyright (C) 2016 Travis Scrimshaw <tscrimsh@umn.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  http://www.gnu.org/licenses/
#******************************************************************************

from sage.categories.commutative_algebras import CommutativeAlgebras
from sage.categories.topological_spaces import TopologicalSpaces
from sage.manifolds.scalarfield import ScalarField
from sage.misc.cachefunc import cached_method
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.symbolic.ring import SR, SymbolicRing


class ScalarFieldAlgebra(UniqueRepresentation, Parent):
    r"""
    Commutative algebra of scalar fields on a topological manifold.

    If `M` is a topological manifold over a topological field `K`, the
    commutative algebra of scalar fields on `M` is the set `C^0(M)` of all
    continuous maps `M \to K`. The set `C^0(M)` is an algebra over `K`,
    whose ring product is the pointwise multiplication of `K`-valued
    functions, which is clearly commutative.

    If `K = \RR` or `K = \CC`, the field `K` over which the
    algebra `C^0(M)` is constructed is represented by the :class:`Symbolic
    Ring <sage.symbolic.ring.SymbolicRing>` ``SR``, since there is no exact
    representation of `\RR` nor `\CC`.

    INPUT:

    - ``domain`` -- the topological manifold `M` on which the scalar fields
      are defined

    EXAMPLES:

    Algebras of scalar fields on the sphere `S^2` and on some open
    subsets of it::

        sage: M = Manifold(2, 'M', structure='topological') # the 2-dimensional sphere S^2
        sage: U = M.open_subset('U')  # complement of the North pole
        sage: c_xy.<x,y> = U.chart()  # stereographic coordinates from the North pole
        sage: V = M.open_subset('V')  # complement of the South pole
        sage: c_uv.<u,v> = V.chart()  # stereographic coordinates from the South pole
        sage: M.declare_union(U,V)    # S^2 is the union of U and V
        sage: xy_to_uv = c_xy.transition_map(c_uv, (x/(x^2+y^2), y/(x^2+y^2)),
        ....:                                intersection_name='W',
        ....:                                restrictions1= x^2+y^2!=0,
        ....:                                restrictions2= u^2+v^2!=0)
        sage: uv_to_xy = xy_to_uv.inverse()
        sage: CM = M.scalar_field_algebra(); CM
        Algebra of scalar fields on the 2-dimensional topological manifold M
        sage: W = U.intersection(V)  # S^2 minus the two poles
        sage: CW = W.scalar_field_algebra(); CW
        Algebra of scalar fields on the Open subset W of the
         2-dimensional topological manifold M

    `C^0(M)` and `C^0(W)` belong to the category of commutative
    algebras over `\RR` (represented here by
    :class:`~sage.symbolic.ring.SymbolicRing`)::

        sage: CM.category()
        Join of Category of commutative algebras over Symbolic Ring and Category of homsets of topological spaces
        sage: CM.base_ring()
        Symbolic Ring
        sage: CW.category()
        Join of Category of commutative algebras over Symbolic Ring and Category of homsets of topological spaces
        sage: CW.base_ring()
        Symbolic Ring

    The elements of `C^0(M)` are scalar fields on `M`::

        sage: CM.an_element()
        Scalar field on the 2-dimensional topological manifold M
        sage: CM.an_element().display()  # this sample element is a constant field
        M → ℝ
        on U: (x, y) ↦ 2
        on V: (u, v) ↦ 2

    Those of `C^0(W)` are scalar fields on `W`::

        sage: CW.an_element()
        Scalar field on the Open subset W of the 2-dimensional topological
         manifold M
        sage: CW.an_element().display()  # this sample element is a constant field
        W → ℝ
        (x, y) ↦ 2
        (u, v) ↦ 2

    The zero element::

        sage: CM.zero()
        Scalar field zero on the 2-dimensional topological manifold M
        sage: CM.zero().display()
        zero: M → ℝ
        on U: (x, y) ↦ 0
        on V: (u, v) ↦ 0

    ::

        sage: CW.zero()
        Scalar field zero on the Open subset W of the 2-dimensional
         topological manifold M
        sage: CW.zero().display()
        zero: W → ℝ
           (x, y) ↦ 0
           (u, v) ↦ 0

    The unit element::

        sage: CM.one()
        Scalar field 1 on the 2-dimensional topological manifold M
        sage: CM.one().display()
        1: M → ℝ
        on U: (x, y) ↦ 1
        on V: (u, v) ↦ 1

    ::

        sage: CW.one()
        Scalar field 1 on the Open subset W of the 2-dimensional topological
         manifold M
        sage: CW.one().display()
        1: W → ℝ
          (x, y) ↦ 1
          (u, v) ↦ 1

    A generic element can be constructed by using a dictionary of
    the coordinate expressions defining the scalar field::

        sage: f = CM({c_xy: atan(x^2+y^2), c_uv: pi/2 - atan(u^2+v^2)}); f
        Scalar field on the 2-dimensional topological manifold M
        sage: f.display()
        M → ℝ
        on U: (x, y) ↦ arctan(x^2 + y^2)
        on V: (u, v) ↦ 1/2*pi - arctan(u^2 + v^2)
        sage: f.parent()
        Algebra of scalar fields on the 2-dimensional topological manifold M

    Specific elements can also be constructed in this way::

        sage: CM(0) == CM.zero()
        True
        sage: CM(1) == CM.one()
        True

    Note that the zero scalar field is cached::

        sage: CM(0) is CM.zero()
        True

    Elements can also be constructed by means of the method
    :meth:`~sage.manifolds.manifold.TopologicalManifold.scalar_field` acting
    on the domain (this allows one to set the name of the scalar field at the
    construction)::

        sage: f1 = M.scalar_field({c_xy: atan(x^2+y^2), c_uv: pi/2 - atan(u^2+v^2)},
        ....:                     name='f')
        sage: f1.parent()
        Algebra of scalar fields on the 2-dimensional topological manifold M
        sage: f1 == f
        True
        sage: M.scalar_field(0, chart='all') == CM.zero()
        True

    The algebra `C^0(M)` coerces to `C^0(W)` since `W` is an open
    subset of `M`::

        sage: CW.has_coerce_map_from(CM)
        True

    The reverse is of course false::

        sage: CM.has_coerce_map_from(CW)
        False

    The coercion map is nothing but the restriction to `W` of scalar fields
    on `M`::

        sage: fW = CW(f) ; fW
        Scalar field on the Open subset W of the
         2-dimensional topological manifold M
        sage: fW.display()
        W → ℝ
          (x, y) ↦ arctan(x^2 + y^2)
          (u, v) ↦ 1/2*pi - arctan(u^2 + v^2)

    ::

        sage: CW(CM.one()) == CW.one()
        True

    The coercion map allows for the addition of elements of `C^0(W)`
    with elements of `C^0(M)`, the result being an element of
    `C^0(W)`::

        sage: s = fW + f
        sage: s.parent()
        Algebra of scalar fields on the Open subset W of the
         2-dimensional topological manifold M
        sage: s.display()
        W → ℝ
          (x, y) ↦ 2*arctan(x^2 + y^2)
          (u, v) ↦ pi - 2*arctan(u^2 + v^2)

    Another coercion is that from the Symbolic Ring.
    Since the Symbolic Ring is the base ring for the algebra ``CM``, the
    coercion of a symbolic expression ``s`` is performed by the operation
    ``s*CM.one()``, which invokes the (reflected) multiplication operator.
    If the symbolic expression does not involve any chart coordinate,
    the outcome is a constant scalar field::

        sage: h = CM(pi*sqrt(2)) ; h
        Scalar field on the 2-dimensional topological manifold M
        sage: h.display()
        M → ℝ
        on U: (x, y) ↦ sqrt(2)*pi
        on V: (u, v) ↦ sqrt(2)*pi
        sage: a = var('a')
        sage: h = CM(a); h.display()
        M → ℝ
        on U: (x, y) ↦ a
        on V: (u, v) ↦ a

    If the symbolic expression involves some coordinate of one of the
    manifold's charts, the outcome is initialized only on the chart domain::

        sage: h = CM(a+x); h.display()
        M → ℝ
        on U: (x, y) ↦ a + x
        on W: (u, v) ↦ (a*u^2 + a*v^2 + u)/(u^2 + v^2)
        sage: h = CM(a+u); h.display()
        M → ℝ
        on W: (x, y) ↦ (a*x^2 + a*y^2 + x)/(x^2 + y^2)
        on V: (u, v) ↦ a + u

    If the symbolic expression involves coordinates of different charts,
    the scalar field is created as a Python object, but is not initialized,
    in order to avoid any ambiguity::

        sage: h = CM(x+u); h.display()
        M → ℝ

    TESTS:

    Ring laws::

        sage: h = CM(pi*sqrt(2))
        sage: s = f + h ; s
        Scalar field on the 2-dimensional topological manifold M
        sage: s.display()
        M → ℝ
        on U: (x, y) ↦ sqrt(2)*pi + arctan(x^2 + y^2)
        on V: (u, v) ↦ 1/2*pi*(2*sqrt(2) + 1) - arctan(u^2 + v^2)

    ::

        sage: s = f - h ; s
        Scalar field on the 2-dimensional topological manifold M
        sage: s.display()
        M → ℝ
        on U: (x, y) ↦ -sqrt(2)*pi + arctan(x^2 + y^2)
        on V: (u, v) ↦ -1/2*pi*(2*sqrt(2) - 1) - arctan(u^2 + v^2)

    ::

        sage: s = f*h ; s
        Scalar field on the 2-dimensional topological manifold M
        sage: s.display()
        M → ℝ
        on U: (x, y) ↦ sqrt(2)*pi*arctan(x^2 + y^2)
        on V: (u, v) ↦ 1/2*sqrt(2)*(pi^2 - 2*pi*arctan(u^2 + v^2))

    ::

        sage: s = f/h ; s
        Scalar field on the 2-dimensional topological manifold M
        sage: s.display()
        M → ℝ
        on U: (x, y) ↦ 1/2*sqrt(2)*arctan(x^2 + y^2)/pi
        on V: (u, v) ↦ 1/4*sqrt(2)*(pi - 2*arctan(u^2 + v^2))/pi

    ::

        sage: f*(h+f) == f*h + f*f
        True

    Ring laws with coercion::

        sage: f - fW == CW.zero()
        True
        sage: f/fW == CW.one()
        True
        sage: s = f*fW ; s
        Scalar field on the Open subset W of the 2-dimensional topological
         manifold M
        sage: s.display()
        W → ℝ
        (x, y) ↦ arctan(x^2 + y^2)^2
        (u, v) ↦ 1/4*pi^2 - pi*arctan(u^2 + v^2) + arctan(u^2 + v^2)^2
        sage: s/f == fW
        True

    Multiplication by a real number::

        sage: s = 2*f ; s
        Scalar field on the 2-dimensional topological manifold M
        sage: s.display()
        M → ℝ
        on U: (x, y) ↦ 2*arctan(x^2 + y^2)
        on V: (u, v) ↦ pi - 2*arctan(u^2 + v^2)

    ::

        sage: 0*f == CM.zero()
        True
        sage: 1*f == f
        True
        sage: 2*(f/2) == f
        True
        sage: (f+2*f)/3 == f
        True
        sage: 1/3*(f+2*f) == f
        True

    The Sage test suite for algebras is passed::

        sage: TestSuite(CM).run()

    It is passed also for `C^0(W)`::

        sage: TestSuite(CW).run()
    """

    Element = ScalarField

    def __init__(self, domain):
        r"""
        Construct an algebra of scalar fields.

        TESTS::

            sage: M = Manifold(2, 'M', structure='topological')
            sage: X.<x,y> = M.chart()
            sage: CM = M.scalar_field_algebra(); CM
            Algebra of scalar fields on the 2-dimensional topological
             manifold M
            sage: type(CM)
            <class 'sage.manifolds.scalarfield_algebra.ScalarFieldAlgebra_with_category'>
            sage: type(CM).__base__
            <class 'sage.manifolds.scalarfield_algebra.ScalarFieldAlgebra'>
            sage: TestSuite(CM).run()
        """
        base_field = domain.base_field()
        if domain.base_field_type() in ['real', 'complex']:
            base_field = SR
        Parent.__init__(self, base=base_field,
                        category=CommutativeAlgebras(base_field) & TopologicalSpaces().Homsets())
        self._domain = domain
        self._populate_coercion_lists_()

    #### Methods required for any Parent
    def _element_constructor_(self, coord_expression=None, chart=None,
                              name=None, latex_name=None):
        r"""
        Construct a scalar field.

        INPUT:

        - ``coord_expression`` -- (default: ``None``) element(s) to construct
          the scalar field; this can be either

          - a scalar field defined on a domain that encompass ``self._domain``;
            then ``_element_constructor_`` return the restriction of
            the scalar field to ``self._domain``
          - a dictionary of coordinate expressions in various charts on the
            domain, with the charts as keys
          - a single coordinate expression; if the argument ``chart`` is
            ``'all'``, this expression is set to all the charts defined
            on the open set; otherwise, the expression is set in the
            specific chart provided by the argument ``chart``

        - ``chart`` -- (default: ``None``) chart defining the coordinates used
          in ``coord_expression`` when the latter is a single coordinate
          expression; if none is provided (default), the default chart of the
          open set is assumed. If ``chart=='all'``, ``coord_expression`` is
          assumed to be independent of the chart (constant scalar field).

        - ``name`` -- (default: ``None``) string; name (symbol) given to the
          scalar field

        - ``latex_name`` -- (default: ``None``) string; LaTeX symbol to denote
          the scalar field; if none is provided, the LaTeX symbol is set to
          ``name``

        If ``coord_expression`` is ``None`` or incomplete, coordinate
        expressions can be added after the creation of the object, by means
        of the methods :meth:`add_expr`, :meth:`add_expr_by_continuation` and
        :meth:`set_expr`

        TESTS::

            sage: M = Manifold(2, 'M', structure='topological')
            sage: X.<x,y> = M.chart()
            sage: CM = M.scalar_field_algebra()
            sage: f = CM({X: x+y^2}); f
            Scalar field on the 2-dimensional topological manifold M
            sage: f.display()
            M → ℝ
            (x, y) ↦ y^2 + x
            sage: f = CM({X: x+y^2}, name='f'); f
            Scalar field f on the 2-dimensional topological manifold M
            sage: f.display()
            f: M → ℝ
               (x, y) ↦ y^2 + x
            sage: U = M.open_subset('U', coord_def={X: x>0})
            sage: CU = U.scalar_field_algebra()
            sage: fU = CU(f); fU
            Scalar field f on the Open subset U of the 2-dimensional topological
             manifold M
            sage: fU.display()
            f: U → ℝ
               (x, y) ↦ y^2 + x
        """
        try:
            if coord_expression.is_trivial_zero():
                return self.zero()
            if (coord_expression - 1).is_trivial_zero():
                return self.one()
        except AttributeError:
            if coord_expression == 0:
                return self.zero()
            if coord_expression == 1:
                return self.one()
        if isinstance(coord_expression, ScalarField):
            if self._domain.is_subset(coord_expression._domain):
                # restriction of the scalar field to self._domain:
                return coord_expression.restrict(self._domain)
        else:
            # Anything going wrong here should produce a readable error:
            try:
                # generic constructor:
                resu = self.element_class(self,
                                          coord_expression=coord_expression,
                                          name=name, latex_name=latex_name,
                                          chart=chart)
            except TypeError:
                raise TypeError("cannot convert " +
                                "{} to a scalar ".format(coord_expression) +
                                "field on {}".format(self._domain))
        return resu

    def _an_element_(self):
        r"""
        Construct some element of the algebra.

        TESTS::

            sage: M = Manifold(2, 'M', structure='topological')
            sage: X.<x,y> = M.chart()
            sage: CM = M.scalar_field_algebra()
            sage: f = CM._an_element_(); f
            Scalar field on the 2-dimensional topological manifold M
            sage: f.display()
            M → ℝ
            (x, y) ↦ 2
        """
        return self.element_class(self, coord_expression=2, chart='all')

    def _coerce_map_from_(self, other):
        r"""
        Determine whether coercion to ``self`` exists from ``other``.

        TESTS::

            sage: M = Manifold(2, 'M', structure='topological')
            sage: X.<x,y> = M.chart()
            sage: CM = M.scalar_field_algebra()
            sage: CM._coerce_map_from_(SR)
            True
            sage: SCR = SR.subring(no_variables=True); SCR
            Symbolic Constants Subring
            sage: CM._coerce_map_from_(SCR)
            True
            sage: CM._coerce_map_from_(X.function_ring())
            True
            sage: U = M.open_subset('U', coord_def={X: x>0})
            sage: CU = U.scalar_field_algebra()
            sage: CM._coerce_map_from_(CU)
            False
            sage: CU._coerce_map_from_(CM)
            True
        """
        from sage.manifolds.chart_func import ChartFunctionRing
        if isinstance(other, SymbolicRing):
            return True  # coercion from the base ring (multiplication by the
                         # algebra unit, i.e. self.one())
                         # cf. ScalarField._lmul_() for the implementation of
                         # the coercion map
        if isinstance(other, ScalarFieldAlgebra):
            return self._domain.is_subset(other._domain)
        if isinstance(other, ChartFunctionRing):
            return self._domain.is_subset(other._chart.domain())
        return False

    #### End of methods required for any Parent

    def _repr_(self):
        r"""
        String representation of ``self``.

        TESTS::

            sage: M = Manifold(2, 'M', structure='topological')
            sage: CM = M.scalar_field_algebra()
            sage: CM._repr_()
            'Algebra of scalar fields on the 2-dimensional topological manifold M'
            sage: CM
            Algebra of scalar fields on the 2-dimensional topological manifold M
        """
        return "Algebra of scalar fields on the {}".format(self._domain)

    def _latex_(self):
        r"""
        LaTeX representation of the object.

        TESTS::

            sage: M = Manifold(2, 'M', structure='topological')
            sage: CM = M.scalar_field_algebra()
            sage: CM._latex_()
            'C^0 \\left(M\\right)'
            sage: latex(CM)
            C^0 \left(M\right)
        """
        return r"C^0 \left(" + self._domain._latex_() + r"\right)"

    @cached_method
    def zero(self):
        r"""
        Return the zero element of the algebra.

        This is nothing but the constant scalar field `0` on the manifold,
        where `0` is the zero element of the base field.

        EXAMPLES::

            sage: M = Manifold(2, 'M', structure='topological')
            sage: X.<x,y> = M.chart()
            sage: CM = M.scalar_field_algebra()
            sage: z = CM.zero(); z
            Scalar field zero on the 2-dimensional topological manifold M
            sage: z.display()
            zero: M → ℝ
               (x, y) ↦ 0

        The result is cached::

            sage: CM.zero() is z
            True
        """
        coord_express = {chart: chart.zero_function()
                         for chart in self._domain.atlas()}
        zero = self.element_class(self,
                                  coord_expression=coord_express,
                                  name='zero', latex_name='0')
        zero._is_zero = True
        zero.set_immutable()
        return zero

    @cached_method
    def one(self):
        r"""
        Return the unit element of the algebra.

        This is nothing but the constant scalar field `1` on the manifold,
        where `1` is the unit element of the base field.

        EXAMPLES::

            sage: M = Manifold(2, 'M', structure='topological')
            sage: X.<x,y> = M.chart()
            sage: CM = M.scalar_field_algebra()
            sage: h = CM.one(); h
            Scalar field 1 on the 2-dimensional topological manifold M
            sage: h.display()
            1: M → ℝ
               (x, y) ↦ 1

        The result is cached::

            sage: CM.one() is h
            True
        """
        coord_express = {chart: chart.one_function()
                         for chart in self._domain.atlas()}
        one = self.element_class(self, coord_expression=coord_express,
                                 name='1', latex_name='1')
        one.set_immutable()
        return one



# sage.doctest: needs sage.rings.padics
r"""
Elements of Berkovich space.

:class:`Berkovich_Element` is an abstract parent class for elements of any Berkovich space.

:class:`Berkovich_Element_Cp_Affine` and :class:`Berkovich_Element_Cp_Projective`
implement elements of Berkovich space over `\CC_p` and `P^1(\CC_p)`. Elements are
determined by specific data and fall into one of the four following types:

- Type I points are represented by a center.

- Type II points are represented by a center and a rational power of `p`.

- Type III points are represented by a center and a nonnegative real radius.

- Type IV points are represented by a finite list of centers and a finite list of
  nonnegative radii.

For an exposition of Berkovich space over `\CC_p`, see Chapter 6 of [Ben2019]_. For a more
involved exposition, see Chapter 1 and 2 of [BR2010]_.

AUTHORS:

- Alexander Galarraga (2020-06-22): initial implementation
"""

# *****************************************************************************
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************

import sage.rings.abc
from sage.categories.function_fields import FunctionFields
from sage.misc.latex import latex
from sage.misc.lazy_import import lazy_import
from sage.rings.infinity import Infinity
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.rings.real_mpfr import RR, RealNumber
from sage.schemes.projective.projective_point import (
    SchemeMorphism_point_projective_field,
)
from sage.schemes.projective.projective_space import ProjectiveSpace
from sage.structure.element import Element, Expression

lazy_import('sage.rings.padics.padic_generic_element', 'pAdicGenericElement')
lazy_import('sage.rings.padics.padic_base_generic', 'pAdicBaseGeneric')


class Berkovich_Element(Element):
    """
    The parent class for any element of a Berkovich space.
    """
    pass


class Berkovich_Element_Cp(Berkovich_Element):
    r"""
    The abstract parent class for any element of Berkovich space over `\CC_p`.

    This class should never be instantiated, instead use :class:`Berkovich_Element_Cp_Affine`
    or :class:`Berkovich_Element_Cp_Projective`.

    EXAMPLES::

        sage: B = Berkovich_Cp_Affine(3)
        sage: B(2)
        Type I point centered at 2 + O(3^20)

    ::

        sage: B(0, 1)
        Type II point centered at 0 of radius 3^0
    """

    def __init__(self, parent, center, radius=None, power=None, prec=20, space_type=None, error_check=True):
        """
        Initialization function.

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(5)
            sage: B(4)
            Type I point centered at 4 + O(5^20)
        """
        from sage.rings.polynomial.polynomial_element import Polynomial
        from sage.rings.fraction_field_element import FractionFieldElement_1poly_field
        self._type = None

        # if radius is a list or a tuple, this is a type 4 point
        if isinstance(radius, (list, tuple)):
            if error_check:
                if not isinstance(center, (list, tuple)):
                    raise TypeError("center was passed a list but radius was not a list")
                if len(radius) != len(center):
                    raise ValueError("the same number of centers and radii "
                                     "must be specified to create "
                                     "a type IV point")
            self._center_lst = list(center)
            self._radius_lst = list(radius)
            self._prec = len(self._radius_lst)
            self._center_func = None
            self._radius_func = None
            self._type = 4
            self._radius = None
            self._center = None
            if not error_check:
                return

        elif isinstance(center, Element) and isinstance(radius, Element):
            from sage.rings.polynomial.multi_polynomial import MPolynomial
            if isinstance(center, MPolynomial):
                try:
                    center = center.univariate_polynomial()
                except AttributeError:
                    raise TypeError('center was %s, a multivariable polynomial' % center)

            # check if the radius and the center are functions
            center_func_check = center.parent() in FunctionFields() or \
                isinstance(center, (Polynomial, FractionFieldElement_1poly_field, Expression))
            radius_func_check = radius.parent() in FunctionFields() or \
                isinstance(radius, (Polynomial, FractionFieldElement_1poly_field, Expression))

            if center_func_check:
                # check that both center and radii are supported univariate function
                center_expr_check = False
                radius_expr_check = False
                if error_check:
                    if isinstance(center, Expression):
                        if len(center.variables()) != 1:
                            raise ValueError("an expression with %s " % (len(center.variables())) +
                                             "variables cannot define the centers approximating a type IV point")
                        else:
                            # we do this since .subs is currently buggy for polynomials but not expressions
                            center_expr_check = True
                    if not radius_func_check:
                        raise TypeError("center was passed a function but radius was not a function")
                    if isinstance(radius, Expression):
                        if len(radius.variables()) != 1:
                            raise ValueError("an expression with %s " % (len(radius.variables())) +
                                             "variables cannot define the radii approximating a type IV point")
                        else:
                            radius_expr_check = True
                else:
                    if isinstance(center, Expression):
                        center_expr_check = True
                    if isinstance(radius, Expression):
                        radius_expr_check = True
                self._type = 4
                self._prec = prec
                center_lst = []
                radius_lst = []
                self._center_func = center
                self._radius_func = radius
                if center_expr_check:
                    x = self._center_func.variables()[0]
                if radius_expr_check:
                    y = self._radius_func.variables()[0]
                for i in range(1, self._prec + 1):
                    if center_expr_check:
                        # we use .subs for expressions to avoid deprecation
                        center_lst.append(self._center_func.subs({x: i}))
                    else:
                        # .subs for polynomials is currently buggy
                        center_lst.append(self._center_func(i))
                    if radius_expr_check:
                        radius_lst.append(self._radius_func.subs({y: i}))
                    else:
                        radius_lst.append(self._radius_func(i))
                self._center_lst = center_lst
                self._radius_lst = radius_lst
                self._radius = None
                self._center = None
                if not error_check:
                    return

        if self._type == 4 and error_check:
            if space_type == "projective":
                for i in range(len(self._center_lst)):
                    center = self._center_lst[i]
                    radius = self._radius_lst[i]
                    # make sure the center is a point of projective space and not the point at infinity
                    if not isinstance(center, SchemeMorphism_point_projective_field):
                        try:
                            center = (self._base_space)(center)
                        except (TypeError, ValueError):
                            raise TypeError('could not convert %s to %s' % (center, self._base_space))
                    if self._base_type == 'padic field':
                        if not isinstance(center.scheme().base_ring(), sage.rings.abc.pAdicField):
                            if not isinstance(center.scheme().base_ring(), pAdicBaseGeneric):
                                try:
                                    center = (self._base_space)(center)
                                except (TypeError, ValueError):
                                    raise ValueError("could not convert %s to %s" % (center, self._base_space))
                            else:
                                # center is padic, not but an element of a scheme over a padic field.
                                # we convert to scheme over a padic field
                                center = ProjectiveSpace(center.scheme().base_ring().fraction_field(), 1)(center)
                        if center.scheme().base_ring().prime() != self._p:
                            raise ValueError("center must be an element of " +
                                             "%s not %s" % self._base_space, center.scheme())
                    elif center not in self._base_space:
                        try:
                            center = (self._base_space)(center)
                        except (TypeError, ValueError):
                            raise ValueError('could not convert %s to %s' % (center, self._base_space))
                    if center.scheme().ambient_space() != center.scheme():
                        raise ValueError("the center of a point of Berkovich space over " +
                                         "P^1(Cp(%s)) must be a point of Cp not %s" % (self._p, center.scheme()))
                    if center == (center.scheme())((1, 0)):
                        raise ValueError("the center of a disk approximating a type IV point of Berkovich " +
                                         "space cannot be centered at %s" % ((center.scheme())((1, 0))))
                    # since we are over a field, we can normalize coordinates. all code assumes normalized coordinates
                    center.normalize_coordinates()
                    # make sure the radius coerces into the reals
                    if not isinstance(radius, RealNumber):
                        if isinstance(radius, Expression):
                            radius = RR(radius)
                        elif RR.has_coerce_map_from(radius.parent()):
                            radius = RR(radius)
                        else:
                            raise TypeError("the radius of a disk approximating a type IV point" +
                                            "must coerce into the real numbers, %s does not coerce" % (radius))
                    if i != 0:
                        # check containment for the sequence of disks
                        previous_center = self._center_lst[i - 1]
                        previous_radius = self._radius_lst[i - 1]
                        dist = self._custom_abs(center[0] - previous_center[0])
                        if previous_radius < radius or dist > previous_radius:
                            raise ValueError("sequence of disks does not define a type IV point as " +
                                             "containment is not proper")
                    self._center_lst[i] = center
                    self._radius_lst[i] = radius
                return
            if space_type == "affine":
                for i in range(len(self._center_lst)):
                    center = self._center_lst[i]
                    radius = self._radius_lst[i]
                    if self._base_type == 'padic field':
                        # make sure the center is in Cp
                        if not isinstance(center, pAdicGenericElement):
                            try:
                                center = (self._base_space)(center)
                            except (TypeError, ValueError):
                                raise TypeError("could not convert %s to %s" % (center, self._base_space))
                        elif not isinstance(center.parent(), sage.rings.abc.pAdicField):
                            # center is padic, not but an element of a padic field. we convert to padic field
                            center = (center.parent().fraction_field())(center)
                        if (center.parent()).prime() != self._p:
                            raise ValueError("center in %s, should be in %s") % (center.parent(), self._base_space)
                    # make sure the center is in the appropriate number field
                    elif center.parent() == self._base_space:
                        try:
                            center = (self._base_space)(center)
                        except (TypeError, ValueError):
                            raise ValueError('could not convert %s to %s' % (center, self._base_space))
                    # make sure the radius coerces into the reals
                    if not isinstance(radius, RealNumber):
                        if isinstance(radius, Expression):
                            radius = RR(radius)
                        elif RR.has_coerce_map_from(radius.parent()):
                            radius = RR(radius)
                            self._radius_lst[i] = radius
                        else:
                            raise ValueError("the radius of a disk approximating a type IV point must " +
                                             "coerce into the real numbers, %s does not coerce" % (radius))
                    if i != 0:
                        # check containment for the sequence of disks
                        previous_center = self._center_lst[i - 1]
                        previous_radius = self._radius_lst[i - 1]
                        dist = self._custom_abs(center - previous_center)
                        if previous_radius < radius or dist > previous_radius:
                            raise ValueError("sequence of disks does not define a type IV point as " +
                                             "containment is not proper")
                    self._center_lst[i] = center
                    self._radius_lst[i] = radius
                return
            raise ValueError("bad value %s passed to space_type. Do not initialize  " % (space_type) +
                             "Berkovich_Element_Cp directly")

        # the point must now be type 1, 2, or 3, so we check that the center is of the appropriate type
        if error_check:
            if space_type == "projective":
                if not isinstance(center, SchemeMorphism_point_projective_field):
                    try:
                        center = (self._base_space)(center)
                    except (ValueError, TypeError):
                        raise TypeError("could not convert %s to %s" % (center, self._base_space))
                if self._base_type == 'padic field':
                    if not isinstance(center.scheme().base_ring(), sage.rings.abc.pAdicField):
                        if not isinstance(center.scheme().base_ring(), pAdicBaseGeneric):
                            try:
                                center = (self._base_space)(center)
                            except (TypeError, ValueError):
                                raise ValueError("could not convert %s to %s" % (center, self._base_space))
                        else:
                            # center is padic, not but an element of a scheme over a padic field.
                            # we convert to scheme over a padic field
                            field_scheme = ProjectiveSpace(center.scheme().base_ring().fraction_field(), 1)
                            try:
                                center = field_scheme(center)
                            except (TypeError, ValueError):
                                raise ValueError('could not convert %s to %s' % center, field_scheme)
                    if center.scheme().base_ring().prime() != self._p:
                        raise ValueError("center must be an element of " +
                                         "%s not %s" % self._base_space, center.scheme())
                elif center not in self._base_space:
                    try:
                        center = (self._base_space)(center)
                    except (TypeError, ValueError):
                        raise ValueError('could not convert %s to %s' % (center, self._base_space))
                if center.scheme().ambient_space() is not center.scheme():
                    raise ValueError("the center of a point of projective Berkovich space cannot be " +
                                     "a point of %s" % (center.scheme()))
                # since we are over a field, we normalize coordinates
                center.normalize_coordinates()
            elif space_type == 'affine':
                if self._base_type == 'padic field':
                    # make sure the center is in Cp
                    if not isinstance(center, pAdicGenericElement):
                        try:
                            center = (self._base_space)(center)
                        except (TypeError, ValueError):
                            raise TypeError("could not convert %s to %s" % (center, self._base_space))
                    elif not isinstance(center.parent(), sage.rings.abc.pAdicField):
                        # center is padic, not but an element of a padic field. we convert to padic field
                        center = (center.parent().fraction_field())(center)
                    if (center.parent()).prime() != self._p:
                        raise ValueError("center in %s, should be in %s") % (center.parent(), self._base_space)
                # make sure the center is in the appropriate number field
                elif not (center.parent() == self._base_space):
                    try:
                        center = (self._base_space)(center)
                    except (TypeError, ValueError):
                        raise ValueError('could not convert %s to %s' % (center, self._base_space))
            else:
                raise ValueError("bad value %s passed to space_type. Do not initialize  " % (space_type) +
                                 "Berkovich_Element_Cp directly")

        self._center = center

        # since this point is not type IV, these are None
        self._center_func = None
        self._center_lst = None
        self._radius_lst = None
        self._radius_func = None

        if (radius is None and power is None) or radius == 0:
            self._type = 1
            self._radius = 0
            self._power = None
            return
        # In order to simplify our representation, type II and III points cannot be centered at infinity
        if space_type == "projective":
            # TODO use involution map to allow for infinity to be passed in as center
            if center[1] == 0:
                raise ValueError('type II and III points can not be centered at infinity')
        if power is not None:
            if error_check:
                try:
                    power = QQ(power)
                except TypeError:
                    raise TypeError("power must convert to rationals")
                if radius is not None:
                    if radius != RR(self._p**power):
                        raise ValueError("conflicting inputs for power and radius")
            self._power = power
            self._radius = RR(self._p**power)
            self._type = 2
            return
        if radius is not None:
            if isinstance(radius, Expression):
                try:
                    power = QQ(radius.log(self._p).expand_log())
                except TypeError:
                    pass
                try:
                    radius = RR(radius)
                    self._radius = radius
                except TypeError:
                    if len(radius.variables()) == 1:
                        raise ValueError('radius univariate function but center is constant. ' +
                                         'this does not define a type IV point')
                    raise TypeError("symbolic radius must be a real number")
            if (not isinstance(radius, RealNumber)) and power is None:
                if RR.has_coerce_map_from(radius.parent()):
                    self._radius = RR(radius)
                else:
                    raise TypeError("radius must coerce into real numbers")
            else:
                self._radius = radius
            if power is not None:
                self._power = power
                self._type = 2
                return
            power = RR(radius.log(self._p))
            if power.is_integer():
                self._power = QQ(power)
                self._type = 2
            else:
                self._type = 3
                self._power = power
            return

        raise ValueError('unknown error constructing point of Berkovich space over Cp')

    def _custom_abs(self, x):
        """
        Return the absolute value of ``x`` with respect to the norm on ``Cp``.

        Used to simplify code, as ``x`` may be a point of a number field
        or a `p`-adic field.

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(QQ, 3)
            sage: Q1 = B(9)
            sage: Q1._custom_abs(Q1.center())
            1/9

        ::

            sage: B = Berkovich_Cp_Affine(Qp(3))
            sage: Q1 = B(9)
            sage: Q1._custom_abs(Q1.center())
            1/9
        """
        if self._base_type == 'padic field':
            return x.abs()
        if x.valuation(self._ideal) == Infinity:
            return 0
        if self._ideal in QQ:
            return self.prime()**(-x.valuation(self._ideal))
        return self.prime()**(-x.valuation(self._ideal) / self._ideal.absolute_ramification_index())

    def center_function(self):
        """
        Return the function defining the centers of disks in the approximation.

        Not defined unless this point is a type IV point created by using
        a univariate function to compute centers.

        OUTPUT: a univariate function

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(5)
            sage: L.<t> = PolynomialRing(Qp(5))
            sage: T = FractionField(L)
            sage: f = T(1/t)
            sage: R.<x> = RR[]
            sage: Y = FractionField(R)
            sage: g = (40*pi)/x                                                         # needs sage.symbolic
            sage: Q1 = B(f, g)                                                          # needs sage.symbolic
            sage: Q1.center_function()                                                  # needs sage.symbolic
            (1 + O(5^20))/((1 + O(5^20))*t)
        """
        if self.type_of_point() != 4:
            raise ValueError('center_function not defined for points which are not type IV')
        if self._center_func is None:
            raise ValueError('this type IV point does not have a center function')
        return self._center_func

    def radius_function(self):
        """
        Return the function defining the radii of disks in the approximation.

        Not defined unless this point is a type IV point created by using
        a univariate function to compute radii.

        OUTPUT: a univariate function

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(5)
            sage: L.<t> = PolynomialRing(Qp(5))
            sage: T = FractionField(L)
            sage: f = T(1/t)
            sage: R.<x> = RR[]
            sage: Y = FractionField(R)
            sage: g = (40*pi)/x                                                         # needs sage.symbolic
            sage: Q1 = B(f, g)                                                          # needs sage.symbolic
            sage: Q1.radius_function()                                                  # needs sage.symbolic
            40.0000000000000*pi/x
        """
        if self.type_of_point() != 4:
            raise ValueError('center_function not defined for points which are not type IV')
        if self._radius_func is None:
            raise ValueError('this type IV point does not have a radius function')
        return self._radius_func

    def precision(self):
        """
        Return the precision of a type IV point.

        This integer is the number of disks used in the approximation of the type IV point.
        Not defined for type I, II, or III points.

        OUTPUT: integer

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(Qp(3))
            sage: d = B([2, 2, 2], [1.761, 1.123, 1.112])
            sage: d.precision()
            3

        TESTS::

            sage: d.precision == d.prec
            True
        """
        if self._type in [1, 2, 3]:
            raise AttributeError("type I, II, and III points do not have a precision")
        return self._prec

    prec = precision

    def ideal(self):
        r"""
        The ideal which defines an embedding of the ``base_ring`` into `\CC_p`.

        If this Berkovich space is backed by a `p`-adic field, then an embedding is
        already specified, and this returns ``None``.

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(QQ, 3)
            sage: B(0).ideal()
            3

            ::

            sage: B = Berkovich_Cp_Projective(3)
            sage: B(0).ideal()
        """
        return self.parent().ideal()

    def power(self):
        r"""
        The power of ``p`` such that `p^\text{power} = \text{radius}`.

        For type II points, always in `\QQ`. For type III points,
        a real number. Not defined for type I or IV points.

        OUTPUT:

        - A rational for type II points.
        - A real number for type III points.

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(1, 9)
            sage: Q1.power()
            2

        ::

            sage: Q2 = B(1, 4)
            sage: Q2.power()
            1.26185950714291
        """
        if self._type in [1, 4]:
            raise AttributeError("type I and IV points do not have a power")
        return self._power

    def radius(self):
        r"""
        Radius of the corresponding disk (or sequence of disks) in `\CC_p`.

        OUTPUT:

        - A nonnegative real number for type I, II, or III points.
        - A list of nonnegative real numbers for type IV points.

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(1, 2/5)
            sage: Q1.radius()
            0.400000000000000

        ::

            sage: d = B([2, 2, 2], [1.761, 1.123, 1.112])
            sage: d.radius()
            [1.76100000000000, 1.12300000000000, 1.11200000000000]
        """
        if self._type == 4:
            return self._radius_lst
        return self._radius

    def diameter(self, basepoint=Infinity):
        r"""
        Generalized diameter function on Berkovich space.

        If the basepoint is infinity, the diameter is equal to
        the limit of the radii of the corresponding disks in `\CC_p`.

        If the basepoint is not infinity, the diameter
        is the Hsia kernel of this point with itself at
        basepoint ``basepoint``.

        INPUT:

        - ``basepoint`` -- (default: ``Infinity``) a point of the
          same Berkovich space as this point

        OUTPUT: a real number

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(3)
            sage: Q1.diameter()
            0

        ::

            sage: Q2 = B(1/2, 9)
            sage: Q2.diameter()
            9.00000000000000

        The diameter of a type IV point is the limit of the radii::

            sage: R.<x> = PolynomialRing(Qp(3))
            sage: f = R(2)
            sage: S.<y> = PolynomialRing(RR)
            sage: S = FractionField(S)
            sage: g = (y+1)/y
            sage: B(f,g).diameter()
            1.0

        ::

            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(1/81, 1)
            sage: Q2 = B(1/3)
            sage: Q1.diameter(Q2)
            0.00137174211248285

        ::

            sage: Q2.diameter(Q2)
            +infinity
        """
        if basepoint == Infinity:
            if self._type == 4:
                if self._radius_func is None:
                    return self._radius_lst[-1]
                from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
                R = PolynomialRing(QQ, names='x')
                x = R.gens()[0]
                if isinstance(self._radius_func, Expression):
                    radius_func_variable = self._radius_func.variables()[0]
                    radius_expr = self._radius_func.subs({radius_func_variable: x})
                else:
                    radius_expr = self._radius_func(x)
                    from sage.symbolic.ring import SymbolicRing as SR
                    radius_expr = SR(RR)(radius_expr)
                return radius_expr.limit(x='oo')
            return self._radius
        if not isinstance(basepoint, Berkovich_Element_Cp):
            raise TypeError('basepoint must be a point of Berkovich space, not %s' % basepoint)
        if basepoint.parent() != self.parent():
            raise ValueError('basepoint must be a point of the same Berkovich space')
        return self.Hsia_kernel(self, basepoint)

    def path_distance_metric(self, other):
        r"""
        Return the path distance metric distance between this point and ``other``.

        Also referred to as the hyperbolic metric, or the big metric.

        On the set of type II, III and IV points, the path distance metric
        is a metric. Following Baker and Rumely, we extend
        the path distance metric to type I points `x`, `y` by `\rho(x,x) = 0` and `\rho(x,y) =
        \infty`. See [BR2010]_.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point

        OUTPUT: a finite or infinite real number

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(1/4, 4)
            sage: Q2 = B(1/4, 6)
            sage: Q1.path_distance_metric(Q2)
            0.369070246428542

        ::

            sage: Q3 = B(1)
            sage: Q3.path_distance_metric(Q1)
            +infinity

        ::

            sage: Q3.path_distance_metric(Q3)
            0
        """
        if not isinstance(other, type(self)):
            raise TypeError('other must be a point of Berkovich space. other was %s' % other)
        if self.parent() != other.parent():
            raise ValueError("other must be a point of the same Berkovich space")
        if self.type_of_point() == 1 or other.type_of_point() == 1:
            if self == other:
                return 0
            return RR(Infinity)
        return 2 * self.join(other).diameter().log(self.prime()) \
            - self.diameter().log(self.prime()) \
            - other.diameter().log(other.prime())

    big_metric = path_distance_metric

    hyperbolic_metric = path_distance_metric

    def Hsia_kernel(self, other, basepoint):
        """
        The Hsia kernel of this point and ``other``,
        with basepoint ``basepoint``.

        The Hsia kernel with arbitrary basepoint
        is a generalization of the Hsia kernel at infinity.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point
        - ``basepoint`` -- a point of the same Berkovich space as this point

        OUTPUT: a finite or infinite real number

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(2, 9)
            sage: Q2 = B(1/27, 1/27)
            sage: Q3 = B(1, 1/3)
            sage: Q1.Hsia_kernel(Q2, Q3)
            0.111111111111111

        ::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(2, 9)
            sage: Q2 = B(1/2)
            sage: Q3 = B(1/2)
            sage: Q1.Hsia_kernel(Q2, Q3)
            +infinity
        """
        if not isinstance(other, type(self)):
            raise TypeError('other must be a point of Berkovich space. other was %s' % other)
        if self.parent() != other.parent():
            raise ValueError("other must be a point of the same Berkovich space")
        if not isinstance(basepoint, type(self)):
            raise TypeError('basepoint must be a point of Berkovich space. basepoint was %s' % basepoint)
        if basepoint.parent() != self.parent():
            raise ValueError("basepoint must be a point of the same Berkovich space")
        if basepoint.type_of_point() == 1:
            if self == basepoint or other == basepoint:
                return RR(Infinity)
        return self.spherical_kernel(other) / \
            (self.spherical_kernel(basepoint) * other.spherical_kernel(basepoint))

    def small_metric(self, other):
        r"""
        Return the small metric distance between this point and ``other``.

        The small metric is an extension of twice
        the spherical distance on `P^1(\CC_p)`.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point

        OUTPUT: a real number

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(1/4, 4)
            sage: Q2 = B(1/4, 6)
            sage: Q1.small_metric(Q2)
            0.0833333333333333

        ::

            sage: B = Berkovich_Cp_Projective(QQ, 5)
            sage: Q1 = B(0, 1)
            sage: Q2 = B(99)
            sage: Q1.small_metric(Q2)
            1.00000000000000

        ::

            sage: Q3 = B(1/4, 4)
            sage: Q3.small_metric(Q2)
            1.75000000000000

        ::

            sage: Q2.small_metric(Q3)
            1.75000000000000
        """
        if not isinstance(other, Berkovich_Element_Cp):
            raise TypeError('other must be a point of affine Berkovich space. other was %s' % other)
        if self.parent() != other.parent():
            raise ValueError('other must be a point of the same Berkovich space')
        gauss = self.parent()(RR(0), RR(1))
        g_greater_than_s = gauss.gt(self)
        g_greater_than_o = gauss.gt(other)
        if g_greater_than_s and g_greater_than_o:
            return 2 * self.join(other, gauss).diameter() - self.diameter() - other.diameter()
        if not g_greater_than_s:
            new_self = self.involution_map()
        else:
            new_self = self
        if not g_greater_than_o:
            new_other = other.involution_map()
        else:
            new_other = other
        return 2 * new_self.join(new_other, gauss).diameter() \
            - new_self.diameter() - new_other.diameter()

    def potential_kernel(self, other, basepoint):
        """
        The potential kernel of this point with ``other``,
        with basepoint ``basepoint``.

        The potential kernel is the hyperbolic distance
        between ``basepoint`` and the join of this point
        with ``other`` relative to ``basepoint``.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point
        - ``basepoint`` -- a point of the same Berkovich space as this point

        OUTPUT: a finite or infinite real number

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(27, 1)
            sage: Q2 = B(1/3, 2)
            sage: Q3 = B(1/9, 1/2)
            sage: Q3.potential_kernel(Q1, Q2)
            0.369070246428543

        ::

            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(27, 1)
            sage: Q2 = B(1/3, 2)
            sage: Q3 = B(1/9, 1/2)
            sage: Q3.potential_kernel(Q1, Q2)
            0.369070246428543
        """
        if not isinstance(other, type(self)):
            raise TypeError('other must be a point of a Berkovich space, not %s' % other)
        if other.parent() != self.parent():
            raise ValueError('other must be a point of the same Berkovich space')
        if not isinstance(basepoint, type(self)):
            raise TypeError('basepoint must be a point of Berkovich line, not %s' % basepoint)
        if basepoint.parent() != self.parent():
            raise ValueError('basepoint must be a point of the same Berkovich space')
        return basepoint.path_distance_metric(self.join(other, basepoint))

    def spherical_kernel(self, other):
        r"""
        The spherical kernel of this point with ``other``.

        The spherical kernel is one possible extension of the spherical
        distance on `P^1(\CC_p)` to the projective Berkovich line.
        See [BR2010]_ for details.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point

        OUTPUT: a real number

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(2, 2)
            sage: Q2 = B(1/9, 1)
            sage: Q1.spherical_kernel(Q2)
            0.500000000000000

        ::

            sage: Q3 = B(2)
            sage: Q3.spherical_kernel(Q3)
            0
        """
        if not isinstance(other, type(self)):
            raise TypeError('other must be a point of Berkovich space, not %s' % other)
        if other.parent() != self.parent():
            raise ValueError('other must be a point of the same Berkovich space')
        gauss_point = self.parent()(ZZ(0), ZZ(1))
        w = self.join(other, gauss_point)
        dist = gauss_point.path_distance_metric(w)
        if dist == Infinity:
            return 0
        return self.prime()**(-dist)

    def Hsia_kernel_infinity(self, other):
        r"""
        Return the Hsia kernel at infinity of this point with ``other``.

        The Hsia kernel at infinity is the natural extension of the
        absolute value on `\CC_p` to Berkovich space.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point

        OUTPUT: a real number

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(Qp(3))
            sage: Q1 = B(1/4, 4)
            sage: Q2 = B(1/4, 6)
            sage: Q1.Hsia_kernel_infinity(Q2)
            6.00000000000000

        ::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: A.<a> = NumberField(x^3 + 20)
            sage: ideal = A.ideal(-1/2*a^2 + a - 3)
            sage: B = Berkovich_Cp_Projective(A, ideal)
            sage: Q1 = B(4)
            sage: Q2 = B(0, 1.5)
            sage: Q1.Hsia_kernel_infinity(Q2)
            1.50000000000000
        """
        return self.join(other).diameter()

    def center(self):
        r"""
        Return the center of the corresponding disk (or sequence of disks)
        in `\CC_p`.

        OUTPUT: an element of the ``base`` of the parent Berkovich space

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(3)
            sage: B(3, 1).center()
            3 + O(3^21)

        ::

            sage: C = Berkovich_Cp_Projective(3)
            sage: C(3, 1).center()
            (3 + O(3^21) : 1 + O(3^20))

        ::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: A.<a> = NumberField(x^3 + 20)
            sage: ideal = A.ideal(-1/2*a^2 + a - 3)
            sage: B = Berkovich_Cp_Projective(A, ideal)
            sage: B(a^2 + 4).center()
            (a^2 + 4 : 1)
        """
        if self._type == 4:
            return self._center_lst
        return self._center

    def type_of_point(self):
        r"""
        Return the type of this point of Berkovich space over `\CC_p`.

        OUTPUT: integer between 1 and 4 inclusive

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(3)
            sage: B(1).type_of_point()
            1

        ::

            sage: B(0, 1).type_of_point()
            2
        """
        return ZZ(self._type)

    def prime(self):
        """
        The residue characteristic of the parent.

        OUTPUT: a prime integer

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(3)
            sage: B(1).prime()
            3
        """
        return ZZ(self._p)

    def __ne__(self, other):
        """
        Non-equality operator.

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(3, 3**(1/2))
            sage: Q2 = B(3, RR(3**(1/2)))
            sage: Q1 != Q2
            False
        """
        return not (self == other)

    def _repr_(self):
        """
        String representation of this point.

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: B(2, 1)
            Type II point centered at (2 + O(3^20) : 1 + O(3^20)) of radius 3^0
        """
        if self._type == 1:
            return "Type I point centered at " + format(self._center)
        if self._type == 2:
            return "Type II point centered at " \
                + format(self._center) \
                + " of radius %s^%s" % (self._p, self._power)
        if self._type == 3:
            return "Type III point centered at " \
                + format(self._center) + " of radius " \
                + format(self._radius)
        if self._center_func is not None and self._radius_func is not None:
            return "Type IV point of precision %s " % self._prec + \
                "with centers given by %s and radii given by %s"\
                % (self._center_func, self._radius_func)
        return "Type IV point of precision %s, approximated " % self._prec + \
            "by disks centered at %s ... with radii %s ..." \
            % (self._center_lst[:min(self._prec, 2)], self._radius_lst[:min(self._prec, 2)])

    def _latex_(self):
        r"""
        LaTeX representation of this point.

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: latex(B(2, 1))
            \text{type 2 Point of } \text{Projective Berkovich line over }
            \Bold{C}_{3} \text{equivalent to the disk centered at
            (2 + O(3^20) : 1 + O(3^20)) of radius 1.00000000000000 in } \Bold{C}_3
        """
        if self._type == 1:
            text = r"the point %s of } \Bold{C}_%s" % (self._center, self._p)
        elif self._type in [2, 3]:
            text = r"the disk centered at %s of radius %s in } \Bold{C}_%s" \
                % (self._center, self._radius, self._p)
        else:
            text = "the sequence of disks with centers %s } " % self._center_lst[:2] + \
                r"\ldots \text{ and radii %s } \ldots" % self._radius_lst[:2]
        return r"\text{type %s Point of }" % (self._type) \
            + latex(self.parent()) + r"\text{equivalent to " + text


class Berkovich_Element_Cp_Affine(Berkovich_Element_Cp):
    r"""
    Element class of the Berkovich affine line over `\CC_p`.

    Elements are categorized into four types, represented by specific data:

    - Type I points are represented by a center in the ``base`` of the parent Berkovich space,
      which is `\QQ_p`, a finite extension of `\QQ_p`, or a number field.

    - Type II points are represented by a center in the ``base`` of the parent Berkovich space,
      and a rational power of `p`.

    - Type III points are represented by a center in the ``base`` of the parent Berkovich space,
      and a radius, a real number in `[0,\infty)`.

    - Type IV points are represented by a finite list of centers in the ``base`` of the parent
      Berkovich space and a finite list of radii in `[0,\infty)`. Type IV points can be created
      from univariate functions, allowing for arbitrary precision.

    INPUT:

    - ``center`` -- for type I, II, and III points, the center of the
      corresponding disk in `\CC_p`. If the parent Berkovich space was created using a number field
      `K`, then ``center`` must be an element of `K`. Otherwise, ``center`` must be an element of a
      `p`-adic field. For type IV points, can be a list of centers used to approximate the point or a
      univariate function that computes the centers (computation starts at 1).

    - ``radius`` -- (optional) For type I, II, and III points, the radius of the
      corresponding disk in ``Cp``. Must coerce into the real numbers. For type IV points,
      can be a list of radii used to approximate the point or a univariate function that
      computes the radii (computation starts at 1).

    - ``power`` -- (optional) Rational number. Used for constructing type II points; specifies
      the power of ``p`` such that `p^\text{power}` = radius

    - ``prec`` -- (default: 20) the number of disks to be used to approximate a type IV point

    - ``error_check`` -- boolean (default: ``True``); if error checking should be run on input. If
      input is correctly formatted, can be set to ``False`` for better performance.
      WARNING: with error check set to ``False``, any error in the input will lead to
      incorrect results.

    EXAMPLES:

    Type I points can be created by specifying the corresponding point of ``Cp``::

        sage: B = Berkovich_Cp_Affine(Qp(3))
        sage: B(4)
        Type I point centered at 1 + 3 + O(3^20)

    The center of a point can be an element of a finite extension of ``Qp``::

        sage: A.<t> = Qq(27)
        sage: B(1 + t)
        Type I point centered at (t + 1) + O(3^20)

    Type II and III points can be created by specifying a center and a radius::

        sage: B(2, 3**(1/2))                                                            # needs sage.symbolic
        Type II point centered at 2 + O(3^20) of radius 3^1/2

    ::

        sage: B(2, 1.6)
        Type III point centered at 2 + O(3^20) of radius 1.60000000000000

    Some type II points may be mistaken for type III points::

        sage: B(3, 3**0.5)                      # not tested
        Type III point centered at 3 + O(3^21) of radius 1.73205080756888

    To avoid these errors, specify the power instead of the radius::

        sage: B(3, power=RR(1/100000))
        Type II point centered at 3 + O(3^21) of radius 3^1/100000

    Type IV points can be constructed in a number of ways, the first being
    from a list of centers and radii used to approximate the point::

        sage: B([Qp(3)(2), Qp(3)(2), Qp(3)(2)], [1.761, 1.123, 1.112])
        Type IV point of precision 3, approximated by disks centered at
        [2 + O(3^20), 2 + O(3^20)] ... with radii [1.76100000000000, 1.12300000000000] ...

    Type IV points can be constructed from univariate functions, with arbitrary precision::

        sage: A.<t> = Qq(27)
        sage: R.<x> = PolynomialRing(A)
        sage: f = (1 + t)^2*x
        sage: S.<y> = PolynomialRing(RR)
        sage: S = FractionField(S)
        sage: g = (y + 1)/y
        sage: d = B(f, g, prec=100); d
        Type IV point of precision 100 with centers given by
        ((t^2 + 2*t + 1) + O(3^20))*x and radii given by (y + 1.00000000000000)/y

    For increased performance, ``error_check`` can be set to ``False``. WARNING: with error check set
    to ``False``, any error in the input will lead to incorrect results::

        sage: B(f, g, prec=100, error_check=False)
        Type IV point of precision 100 with centers given by
        ((t^2 + 2*t + 1) + O(3^20))*x and radii given by (y + 1.00000000000000)/y

    When creating a Berkovich space backed by a number field, points can be created similarly::

        sage: # needs sage.rings.number_field
        sage: R.<x> = QQ[]
        sage: A.<a> = NumberField(x^3 + 20)
        sage: ideal = A.prime_above(3)
        sage: B = Berkovich_Cp_Projective(A, ideal)
        sage: Q1 = B(a); Q1
        Type I point centered at (a : 1)

    ::

        sage: B(a + 1, 3)                                                               # needs sage.rings.number_field
        Type II point centered at (a + 1 : 1) of radius 3^1

    TESTS::

        sage: A = Berkovich_Cp_Affine(3)
        sage: Q1 = A(3, 1); Q1
        Type II point centered at 3 + O(3^21) of radius 3^0
        sage: Q2 = A(2.5, 1); Q2
        Type II point centered at 1 + 2*3 + 3^2 + 3^3 + 3^4 + 3^5 + 3^6 + 3^7 +
        3^8 + 3^9 + 3^10 + 3^11 + 3^12 + 3^13 + 3^14 + 3^15 + 3^16 + 3^17 +
        3^18 + 3^19 + O(3^20) of radius 3^0
        sage: Q5 = A(3, 0); Q5
        Type I point centered at 3 + O(3^21)
        sage: A(Zp(3)(2), 2).center().parent() == A(Qp(3)(2), 2).center().parent()
        True
        sage: Q1 == Q2
        True
        sage: Q1 == Q5
        False
        sage: Q3 = A(Qp(3)(3), power=0, error_check=False); Q3
        Type II point centered at 3 + O(3^21) of radius 3^0
        sage: Q4 = A(3, 3**0); Q4
        Type II point centered at 3 + O(3^21) of radius 3^0
        sage: Q5 = A(3, power=1/2); Q5
        Type II point centered at 3 + O(3^21) of radius 3^1/2
        sage: Q6 = A(3, RR(3**(1/2))); Q6                                               # needs sage.symbolic
        Type III point centered at 3 + O(3^21) of radius 1.73205080756888
        sage: Q5 == Q6                                                                  # needs sage.symbolic
        True

        sage: k = Qp(5)
        sage: R.<x> = k[]
        sage: l.<w> = k.extension(x^2 - 5)
        sage: B = Berkovich_Cp_Affine(5)
        sage: B(w, power=1)
        Type II point centered at w + O(w^41) of radius 5^1

        sage: TestSuite(Q5).run()
    """

    def __init__(self, parent, center, radius=None, power=None, prec=20, error_check=True):
        """
        Initialization function.

        EXAMPLES::

            sage: A = Berkovich_Cp_Affine(17)
            sage: A(5, 1)
            Type II point centered at 5 + O(17^20) of radius 17^0
        """
        # we call Berkovich_Element_Cp constructor which is shared with projective Berkovich space
        # unless we are passed a point of projective Berkovich space
        Element.__init__(self, parent)
        self._p = parent.prime()
        self._base_space = parent.base()
        self._base_type = parent._base_type
        self._ideal = parent._ideal

        # if this is a point of projective Berkovich space, we raise an error
        if isinstance(center, Berkovich_Element_Cp_Projective):
            raise TypeError('use as_affine_point to convert to affine Berkovich space')

        Berkovich_Element_Cp.__init__(self, parent=parent, center=center, radius=radius, power=power,
                                      prec=prec, space_type='affine', error_check=error_check)

    def as_projective_point(self):
        r"""
        Return the corresponding point of projective Berkovich space.

        We identify affine Berkovich space with the subset `P^1_{\text{Berk}}(C_p) - \{(1 : 0)\}`.

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(5)
            sage: B(5).as_projective_point()
            Type I point centered at (5 + O(5^21) : 1 + O(5^20))

        ::

            sage: B(0, 1).as_projective_point()
            Type II point centered at (0 : 1 + O(5^20)) of radius 5^0

        ::

            sage: L.<t> = PolynomialRing(Qp(5))
            sage: T = FractionField(L)
            sage: f = T(1/t)
            sage: R.<x> = RR[]
            sage: Y = FractionField(R)
            sage: g = (40*pi)/x                                                         # needs sage.symbolic
            sage: Q2 = B(f, g)                                                          # needs sage.symbolic
            sage: Q2.as_projective_point()                                              # needs sage.symbolic
            Type IV point of precision 20 with centers given by (1 + O(5^20))/((1 + O(5^20))*t)
            and radii given by 40.0000000000000*pi/x
        """
        from sage.schemes.berkovich.berkovich_space import Berkovich_Cp_Projective
        new_space = Berkovich_Cp_Projective(self.parent().base_ring(), self.parent().ideal())
        if self.type_of_point() == 1:
            return new_space(self.center())
        if self.type_of_point() == 2:
            return new_space(self.center(), power=self.power())
        if self.type_of_point() == 3:
            return new_space(self.center(), self.radius())
        if self._center_func is None:
            center = self.center()
        else:
            center = self.center_function()
        if self._radius_func is None:
            radius = self.radius()
        else:
            radius = self.radius_function()
        return new_space(center, radius, prec=self.prec())

    def __eq__(self, other):
        """
        Equality operator.

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(1, RR(3**(1/2)))                                               # needs sage.symbolic
            sage: Q2 = B(1, 3**(1/2))                                                   # needs sage.symbolic
            sage: Q1 == Q2                                                              # needs sage.symbolic
            True

        ::

            sage: Q3 = B(1)
            sage: Q4 = B(4)
            sage: Q3 == Q4
            False

        ::

            sage: Q5 = B(1, 4)
            sage: Q1 == Q5                                                              # needs sage.symbolic
            False

        ::

            sage: Q1 == Q3                                                              # needs sage.symbolic
            False
        """
        if other is self:
            return True
        if not isinstance(other, Berkovich_Element_Cp_Affine):
            return False
        if other.parent() != self.parent():
            return False
        stype = self.type_of_point()
        otype = other.type_of_point()
        if stype == otype and stype == 1:
            return self.center() == other.center()
        if stype == otype and stype == 4:
            raise NotImplementedError("Equality for type IV points not yet implemented")
        elif stype in [2, 3] and otype in [2, 3]:
            if self.radius() != other.radius():
                return False
            center_dist = self._custom_abs(self.center() - other.center())
            return center_dist <= self.radius()
        else:
            return False

    def __hash__(self):
        """
        Return the hash of this point.

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(1, RR(3**(1/2)))                                               # needs sage.symbolic
            sage: Q2 = B(1, 3**(1/2))                                                   # needs sage.symbolic
            sage: hash(Q1) == hash(Q2)                                                  # needs sage.symbolic
            True

        ::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: A.<a> = NumberField(x^3 + 20)
            sage: ideal = A.ideal(-1/2*a^2 + a - 3)
            sage: B = Berkovich_Cp_Projective(A, ideal)
            sage: Q1 = B(a^2 + 1, 2)
            sage: Q2 = B(0, 2)
            sage: hash(Q1) == hash(Q2)
            True
        """
        if self.type_of_point() == 1:
            return hash(self.center())
        if self.type_of_point() == 4:
            raise NotImplementedError('hash not defined for type IV points')
        return hash(self.radius())

    def lt(self, other):
        r"""
        Return ``True`` if this point is strictly less than ``other`` in the standard partial order.

        Roughly, the partial order corresponds to containment of
        the corresponding disks in ``Cp``.

        For example, let x and y be points of type II or III.
        If x has center `c_1` and radius `r_1` and y has center
        `c_2` and radius `r_2`, `x < y` if and only if `D(c_1,r_1)`
        is a subset of `D(c_2,r_2)` in `\CC_p`.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point

        OUTPUT:

        - ``True`` -- if this point is less than ``other`` in the standard partial order
        - ``False`` -- otherwise

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(5, 0.5)
            sage: Q2 = B(5, 1)
            sage: Q1.lt(Q2)
            True

        ::

            sage: Q3 = B(1)
            sage: Q1.lt(Q3)
            False

        TESTS::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(5)
            sage: Q1.lt(Q1)
            False

        ::

            sage: Q2 = B([4, 1/3], [5, 1])
            sage: Q1.lt(Q2)
            False

        ::

            sage: Q4 = B(0, 1)
            sage: Q1.lt(Q4)
            True

        ::

            sage: Q2.lt(Q4)
            False
        """
        if not isinstance(other, Berkovich_Element_Cp_Affine):
            raise TypeError('other must be a point of a projective Berkovich space, but was %s' % other)
        if self.parent() != other.parent():
            raise ValueError('other must be a point of the same projective Berkovich space')

        if self == other:
            return False
        if other.type_of_point() in [1, 4]:
            return False

        if self.type_of_point() == 4:
            center = self.center()[-1]
            dist = self._custom_abs(other.center() - center)
            return dist <= other.radius() and self.radius()[-1] <= other.radius()
        dist = self._custom_abs(self.center() - other.center())
        return dist <= other.radius() and self.radius() <= other.radius()

    def gt(self, other):
        r"""
        Return ``True`` if this point is strictly greater than ``other`` in the standard partial order.

        Roughly, the partial order corresponds to containment of
        the corresponding disks in `\CC_p`.

        For example, let x and y be points of type II or III.
        If x has center `c_1` and radius `r_1` and y has center
        `c_2` and radius `r_2`, `x < y` if and only if `D(c_1,r_1)`
        is a subset of `D(c_2,r_2)` in `\CC_p`.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point

        OUTPUT:

        - ``True`` -- if this point is greater than ``other`` in the standard partial order
        - ``False`` -- otherwise

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(QQ, 3)
            sage: Q1 = B(5, 3)
            sage: Q2 = B(5, 1)
            sage: Q1.gt(Q2)
            True

        ::

            sage: Q3 = B(1/27)
            sage: Q1.gt(Q3)
            False

        TESTS::

            sage: B = Berkovich_Cp_Affine(QQ, 3)
            sage: Q1 = B(5)
            sage: Q1.gt(Q1)
            False

        ::

            sage: Q2 = B(0, 1)
            sage: Q1.gt(Q2)
            False

        ::

            sage: Q3 = B([0, 3], [5, 1])
            sage: Q2.gt(Q3)
            True
        """
        if not isinstance(other, Berkovich_Element_Cp_Affine):
            raise TypeError('other must be a point of a projective Berkovich space, but was %s' % other)
        if self.parent() != other.parent():
            raise ValueError('other must be a point of the same projective Berkovich space')

        if self == other:
            return False
        if self.type_of_point() in [1, 4]:
            return False

        if other.type_of_point() == 4:
            center = other.center()[-1]
            dist = self._custom_abs(self.center() - center)
            return dist <= self.radius() and other.radius()[-1] <= self.radius()
        dist = self._custom_abs(self.center() - other.center())
        return dist <= self.radius() and other.radius() <= self.radius()

    def join(self, other, basepoint=Infinity):
        """
        Compute the join of this point and ``other`` with respect to ``basepoint``.

        The join is first point that lies on the intersection
        of the path from this point to ``basepoint`` and the path from ``other`` to
        ``basepoint``.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point
        - ``basepoint`` -- (default: ``Infinity``) a point of the same
          Berkovich space as this point or ``Infinity``

        OUTPUT: a point of the same Berkovich space

        EXAMPLES::

            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(2, 1)
            sage: Q2 = B(2, 2)
            sage: Q1.join(Q2)
            Type III point centered at 2 + O(3^20) of radius 2.00000000000000

        ::

            sage: Q3 = B(5)
            sage: Q3.join(Q1)
            Type II point centered at 2 + 3 + O(3^20) of radius 3^0

        ::

            sage: Q3.join(Q1, basepoint=Q2)
            Type II point centered at 2 + O(3^20) of radius 3^0

        TESTS::

            sage: Q4 = B(1/3**8 + 2, 1)
            sage: Q2.join(Q4, basepoint=Q1)
            Type III point centered at 2 + O(3^20) of radius 2.00000000000000

        ::

            sage: Q5 = B(2, 1/9)
            sage: Q6 = B(1, 1/27)
            sage: Q4.join(Q5, basepoint=Q6)
            Type II point centered at 1 + O(3^20) of radius 3^0

        ::

            sage: Q7 = B(1/27, 1/27)
            sage: Q1.join(Q7, Q2)
            Type III point centered at 2 + O(3^20) of radius 2.00000000000000
        """
        # we error check and then pass to projective space to do the join
        if not isinstance(other, Berkovich_Element_Cp_Affine):
            raise TypeError('other must be a point of affine Berkovich space. other was %s' % other)
        if self.parent() != other.parent():
            raise ValueError('other must be a point of the same affine Berkovich space')
        if self.type_of_point() == 4 or other.type_of_point() == 4:
            raise NotImplementedError("join with type IV points not implemented")

        proj_self = self.as_projective_point()
        proj_other = other.as_projective_point()

        if basepoint == Infinity:
            return proj_self.join(proj_other).as_affine_point()

        if not isinstance(basepoint, Berkovich_Element_Cp_Affine):
            raise TypeError('basepoint must a point of affine Berkovich space. basepoint was %s' % basepoint)
        if basepoint.parent() != self.parent():
            raise ValueError("basepoint must be a point of the same affine Berkovich space")
        if basepoint.type_of_point() == 4:
            raise NotImplementedError("join not implemented for type IV basepoint")
        proj_basepoint = basepoint.as_projective_point()
        return proj_self.join(proj_other, proj_basepoint).as_affine_point()

    def involution_map(self):
        r"""
        Return the image of this point under the involution map.

        The involution map is the extension of the map ``z |-> 1/z``
        on `\CC_p` to Berkovich space.

        For affine Berkovich space, not defined for the type I
        point centered at 0.

        If zero is contained in every disk approximating a type IV point,
        then the image under the involution map is not defined. To avoid
        this error, increase precision.

        OUTPUT: a point of the same Berkovich space

        EXAMPLES:

        The involution map is 1/z on type I points::

            sage: B = Berkovich_Cp_Affine(3)
            sage: Q1 = B(1/2)
            sage: Q1.involution_map()
            Type I point centered at 2 + O(3^20)

        ::

            sage: Q2 = B(0, 1/3)
            sage: Q2.involution_map()
            Type II point centered at 0 of radius 3^1

        ::

            sage: Q3 = B(1/3, 1/3)
            sage: Q3.involution_map()
            Type II point centered at 3 + O(3^21) of radius 3^-3

        TESTS::

            sage: B = Berkovich_Cp_Affine(3)
            sage: B(0).involution_map()
            Traceback (most recent call last):
            ...
            ValueError: involution map not defined on affine type I point centered at 0

        ::

            sage: B(1/81, 1.5).involution_map()
            Type III point centered at 3^4 + O(3^24) of radius 0.000228623685413809

        ::

            sage: B([1, 2], [3, 1]).involution_map()
            Traceback (most recent call last):
            ...
            ValueError: precision of type IV is not high enough to define image

        ::

            sage: B([1/81, 10/81], [10, 9]).involution_map()
            Type IV point of precision 2, approximated by disks centered at [3^4 + O(3^24),
            3^4 + 2*3^6 + 2*3^7 + 2*3^10 + 2*3^11 + 2*3^14 + 2*3^15 + 2*3^18 + 2*3^19 + 2*3^22
            + 2*3^23 + O(3^24)] ... with radii [0.00152415790275873, 0.00137174211248285] ...
        """
        if self.type_of_point() == 1:
            if self.center() == 0:
                raise ValueError("involution map not defined on affine type I point centered at 0")
            return self.parent()(1 / self.center())

        zero = self.parent()(ZZ(0))
        radius = self.radius()

        if self.type_of_point() in [2, 3]:
            zero_contained_in_self = self.gt(zero)
            if zero_contained_in_self:
                if self.type_of_point() == 2:
                    power = self.power()
                    return self.parent()(ZZ(0), power=-power)
                return self.parent()(ZZ(0), RR(1 / radius))
            return self.parent()(1 / self.center(), RR(radius / (self._custom_abs(self.center())**2)))

        new_center_lst = []
        new_radius_lst = []
        for i in range(len(self.center())):
            berk_point = self.parent()(self.center()[i], self.radius()[i])
            zero_check = berk_point.gt(zero)
            if zero_check:
                continue
            else:
                new_center = 1 / self.center()[i]
                new_radius = self.radius()[i] / (self._custom_abs(self.center()[i])**2)
                new_center_lst.append(new_center)
                new_radius_lst.append(new_radius)
        if not new_center_lst:
            raise ValueError('precision of type IV is not high enough to define image')
        return self.parent()(new_center_lst, new_radius_lst, error_check=False)

    def contained_in_interval(self, start, end):
        """
        Check if this point is an element of the interval [``start``, ``end``].

        INPUT:

        - ``start`` -- a point of the same Berkovich space as this point
        - ``end`` -- a point of the same Berkovich space as this point

        OUTPUT:

        - ``True`` if this point is an element of [``start``, ``end``].
        - ``False`` otherwise.

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective((3))
            sage: Q1 = B(2, 1)
            sage: Q2 = B(2, 4)
            sage: Q3 = B(1/3)
            sage: Q2.contained_in_interval(Q1, Q3.join(Q1))
            False

        ::

            sage: Q4 = B(1/81, 1)
            sage: Q2.contained_in_interval(Q1, Q4.join(Q1))
            True
        """
        if not isinstance(start, Berkovich_Element_Cp_Affine):
            raise TypeError("start must be a point of affine Berkovich space. start was %s" % start)
        if start.parent() != self.parent():
            raise ValueError("start must be a point of the same Berkovich space as this point")
        if not isinstance(end, Berkovich_Element_Cp_Affine):
            raise TypeError("end must be a point of affine Berkovich space. end was %s" % end)
        if end.parent() != self.parent():
            raise ValueError("end must be a point of the same Berkovich space as this point")

        proj_self = self.as_projective_point()
        proj_start = start.as_projective_point()
        proj_end = end.as_projective_point()
        return proj_self.contained_in_interval(proj_start, proj_end)


class Berkovich_Element_Cp_Projective(Berkovich_Element_Cp):
    r"""
    Element class of the Berkovich projective line over `\CC_p`.

    Elements are categorized into four types, represented by specific data:

    - Type I points are represented by a center in the ``base`` of the parent Berkovich space,
      which is projective space of dimension 1 over either `\QQ_p`, a finite extension of `\QQ_p`,
      or a number field.

    - Type II points are represented by a center in the ``base`` of the parent Berkovich space,
      and a rational power of `p`.

    - Type III points are represented by a center in the ``base`` of the parent Berkovich space,
      and by a radius, a real number, in `[0,\infty)`.

    - Type IV points are represented by a finite list of centers in the ``base`` of the parent
      Berkovich space and a finite list of radii in `[0,\infty)`.

    The projective Berkovich line is viewed as the one-point compactification of
    the affine Berkovich line. The projective Berkovich line therefore contains
    every point of the affine Berkovich line, along with a type I point centered
    at infinity.

    INPUT:

    - ``center`` -- for type I, II, and III points, the center of the
      corresponding disk in `P^1(\CC_p)`. If the parent Berkovich space was created using a number field
      `K`, then ``center`` can be an element of `P^1(K)`. Otherwise, ``center``
      must be an element of a projective space of dimension 1 over a `p`-adic field.
      For type IV points, can be a list of centers used to approximate the point or a
      univariate function that computes the centers (computation starts at 1).

    - ``radius`` -- (optional) For type I, II, and III points, the radius of the
      corresponding disk in `\CC_p`. Must coerce into the real numbers. For type IV points,
      can be a list of radii used to approximate the point or a univariate function that
      computes the radii (computation starts at 1).

    - ``power`` -- (optional) Rational number. Used for constructing type II points; specifies
      the power of ``p`` such that `p^\text{power}` = radius

    - ``prec`` -- (default: 20) the number of disks to be used to approximate a type IV point

    - ``error_check`` -- boolean (default: ``True``); if error checking should be run on input. If
      input is correctly formatted, can be set to ``False`` for better performance.
      WARNING: with error check set to ``False``, any error in the input will lead to
      incorrect results.

    EXAMPLES:

    Type I points can be created by specifying the corresponding point of `P^1(\CC_p)`::

        sage: S = ProjectiveSpace(Qp(5), 1)
        sage: P = Berkovich_Cp_Projective(S); P
        Projective Berkovich line over Cp(5) of precision 20

    ::

        sage: a = S(0, 1)
        sage: Q1 = P(a); Q1
        Type I point centered at (0 : 1 + O(5^20))

    ::

        sage: Q2 = P((1,0)); Q2
        Type I point centered at (1 + O(5^20) : 0)

    Type II and III points can be created by specifying a center and a radius::

        sage: Q3 = P((0,5), 5**(3/2)); Q3                                               # needs sage.symbolic
        Type II point centered at (0 : 1 + O(5^20)) of radius 5^3/2

    ::

        sage: Q4 = P(0, 3**(3/2)); Q4                                                   # needs sage.symbolic
        Type III point centered at (0 : 1 + O(5^20)) of radius 5.19615242270663

    Type IV points can be created from lists of centers and radii::

        sage: b = S((3,2))  # create centers
        sage: c = S((4,3))
        sage: d = S((2,3))
        sage: L = [b, c, d]
        sage: R = [1.761, 1.123, 1.112]
        sage: Q5 = P(L, R); Q5
        Type IV point of precision 3, approximated by disks centered at
        [(4 + 2*5 + 2*5^2 + 2*5^3 + 2*5^4 + 2*5^5 + 2*5^6 + 2*5^7 + 2*5^8 + 2*5^9 + 2*5^10 +
         2*5^11 + 2*5^12 + 2*5^13 + 2*5^14 + 2*5^15 + 2*5^16 + 2*5^17 + 2*5^18 + 2*5^19 + O(5^20) :
         1 + O(5^20)), (3 + 3*5 + 5^2 + 3*5^3 + 5^4 + 3*5^5 + 5^6 + 3*5^7 + 5^8 + 3*5^9 +
         5^10 + 3*5^11 + 5^12 + 3*5^13 + 5^14 + 3*5^15 + 5^16 + 3*5^17 + 5^18 + 3*5^19 + O(5^20) :
         1 + O(5^20))] ... with radii [1.76100000000000, 1.12300000000000] ...

    Type IV points can also be created from univariate functions. Since the centers of
    the sequence of disks can not be the point at infinity in `P^1(\CC_p)`, only functions
    into `\CC_p` are supported::

        sage: L.<t> = PolynomialRing(Qp(5))
        sage: T = FractionField(L)
        sage: f = T(1/t)
        sage: R.<x> = RR[]
        sage: Y = FractionField(R)
        sage: g = (40*pi)/x                                                             # needs sage.symbolic
        sage: Q6 = P(f, g); Q6                                                          # needs sage.symbolic
        Type IV point of precision 20 with centers given by (1 + O(5^20))/((1 + O(5^20))*t)
         and radii given by 40.0000000000000*pi/x

    TESTS::

        sage: P((1,0), 3)
        Traceback (most recent call last):
        ...
        ValueError: type II and III points can not be centered at infinity

        sage: B = Berkovich_Cp_Projective(3)
        sage: Q1 = B(3)
        sage: TestSuite(Q1).run()
    """

    def __init__(self, parent, center, radius=None, power=None, prec=20, error_check=True):
        """
        Initialization function.

        EXAMPLES::

            sage: S = ProjectiveSpace(Qp(7), 1)
            sage: P = Berkovich_Cp_Projective(S)
            sage: P(0,1)
            Type II point centered at (0 : 1 + O(7^20)) of radius 7^0
        """
        # if we are given a point of Affine Berkovich Space, we do the conversion
        # otherwise we call the Berkovich_Element_Cp constructor with space_type="projective"
        Element.__init__(self, parent)
        self._p = parent.prime()
        self._base_space = parent.base()
        self._base_type = parent._base_type
        self._ideal = parent._ideal

        # conversion from Affine points is handled in this constructor
        if isinstance(center, Berkovich_Element_Cp_Affine):
            raise TypeError('use as_projective_point to convert to projective Berkovich space')

        Berkovich_Element_Cp.__init__(self, parent=parent, center=center, radius=radius, power=power,
                                      prec=prec, space_type='projective', error_check=error_check)

    def as_affine_point(self):
        """
        Return the corresponding affine point after dehomogenizing at infinity.

        OUTPUT: a point of affine Berkovich space

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(5)
            sage: B(5).as_affine_point()
            Type I point centered at 5 + O(5^21)

        ::

            sage: Q = B(0, 1).as_affine_point(); Q
            Type II point centered at 0 of radius 5^0
            sage: Q.parent()
            Affine Berkovich line over Cp(5) of precision 20

        ::

            sage: L.<t> = PolynomialRing(Qp(5))
            sage: T = FractionField(L)
            sage: f = T(1/t)
            sage: R.<x> = RR[]
            sage: Y = FractionField(R)
            sage: g = (40*pi)/x                                                         # needs sage.symbolic
            sage: Q2 = B(f, g)                                                          # needs sage.symbolic
            sage: Q2.as_affine_point()                                                  # needs sage.symbolic
            Type IV point of precision 20 with centers given by (1 + O(5^20))/((1 + O(5^20))*t)
            and radii given by 40.0000000000000*pi/x
        """
        if self.center()[1] == 0:
            raise ValueError('cannot convert infinity to affine Berkovich space')
        from sage.schemes.berkovich.berkovich_space import Berkovich_Cp_Affine
        new_space = Berkovich_Cp_Affine(self.parent().base_ring(), self.parent().ideal())
        if self.type_of_point() in [1, 2, 3]:
            center = self.center()[0]
            if self.type_of_point() == 1:
                return new_space(center)
            if self.type_of_point() == 2:
                return new_space(center, power=self.power())
            if self.type_of_point() == 3:
                return new_space(center, self.radius())
        if self._center_func is None:
            center = [i[0] for i in self.center()]
        else:
            center = self.center_function()
        if self._radius_func is None:
            radius = self.radius()
        else:
            radius = self.radius_function()
        return new_space(center, radius, prec=self.prec())

    def __eq__(self, other):
        """
        Equality operator.

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B([2, 2], RR(3**(1/2)))                                          # needs sage.symbolic
            sage: Q2 = B([1, 1], 3**(1/2))                                              # needs sage.symbolic
            sage: Q1 == Q2                                                              # needs sage.symbolic
            True

        ::

            sage: Q3 = B(1)
            sage: Q4 = B(4)
            sage: Q3 == Q4
            False

        ::

            sage: Q5 = B(1, 4)
            sage: Q1 == Q5                                                              # needs sage.symbolic
            False

        ::

            sage: Q1 == Q3                                                              # needs sage.symbolic
            False
        """
        if other is self:
            return True
        if not isinstance(other, Berkovich_Element_Cp_Projective):
            return False
        if other.parent() != self.parent():
            return False
        stype = self.type_of_point()
        otype = other.type_of_point()
        if stype == otype and stype == 1:
            return self.center() == other.center()
        if stype == otype and stype == 4:
            raise NotImplementedError("equality for type IV points not implemented")
        elif stype in [2, 3] and otype in [2, 3]:
            if self.radius() != other.radius():
                return False
            scent = self.center()[0]
            ocent = other.center()[0]
            center_dist = self._custom_abs(scent - ocent)
            return center_dist <= self.radius()
        else:
            return False

    def __hash__(self):
        """
        Return the hash of this point.

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: P = ProjectiveSpace(B.base_ring(), 1)
            sage: Q1 = B(P.point([2, 2], False), RR(3**(1/2)))                          # needs sage.symbolic
            sage: Q2 = B([1, 1], 3**(1/2))                                              # needs sage.symbolic
            sage: hash(Q1) == hash(Q2)                                                  # needs sage.symbolic
            True

        ::

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: A.<a> = NumberField(x^3 + 20)
            sage: ideal = A.ideal(-1/2*a^2 + a - 3)
            sage: B = Berkovich_Cp_Projective(A, ideal)
            sage: Q1 = B(a^2 + 1, 2)
            sage: Q2 = B(0, 2)
            sage: hash(Q1) == hash(Q2)
            True
        """
        if self.type_of_point() == 1:
            return hash(self.center())
        if self.type_of_point() == 4:
            raise ValueError('hash not defined for type IV points')
        return hash(self.radius())

    def lt(self, other):
        r"""
        Return ``True`` if this point is strictly less than ``other`` in the standard partial order.

        Roughly, the partial order corresponds to containment of
        the corresponding disks in `\CC_p`.

        For example, let x and y be points of type II or III.
        If x has center `c_1` and radius `r_1` and y has center
        `c_2` and radius `r_2`, `x < y` if and only if `D(c_1,r_1)`
        is a subset of `D(c_2,r_2)` in `\CC_p`.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point

        OUTPUT:

        - ``True`` -- if this point is less than ``other`` in the standard partial order
        - ``False`` -- otherwise

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(5, 0.5)
            sage: Q2 = B(5, 1)
            sage: Q1.lt(Q2)
            True

        ::

            sage: Q3 = B(1)
            sage: Q1.lt(Q3)
            False

        TESTS::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(5)
            sage: Q1.lt(Q1)
            False

        ::

            sage: Q2 = B([4, 1/3], [5, 1])
            sage: Q1.lt(Q2)
            False

        ::

            sage: Q3 = B((1,0))
            sage: Q4 = B(0, 1)
            sage: Q3.lt(Q4)
            False

        ::

            sage: Q4.lt(Q3)
            True

        ::

            sage: Q1.lt(Q4)
            True

        ::

            sage: Q2.lt(Q4)
            False
        """
        if not isinstance(other, Berkovich_Element_Cp_Projective):
            raise TypeError('other must be a point of a projective Berkovich space, but was %s' % other)
        if self.parent() != other.parent():
            raise ValueError('other must be a point of the same projective Berkovich space')

        if self == other:
            return False

        # infinity is maximal with respect to the standard partial order
        infinity = self.parent()((1, 0))
        if self == infinity:
            return False
        if other == infinity:
            return True

        if other.type_of_point() in [1, 4]:
            return False
        if self.type_of_point() == 4:
            center = self.center()[-1]
            dist = self._custom_abs(other.center()[0] - center[0])
            return dist <= other.radius() and self.radius()[-1] <= other.radius()
        dist = self._custom_abs(self.center()[0] - other.center()[0])
        return dist <= other.radius() and self.radius() <= other.radius()

    def gt(self, other):
        r"""
        Return ``True`` if this point is strictly greater than ``other`` in the standard partial order.

        Roughly, the partial order corresponds to containment of
        the corresponding disks in `\CC_p`.

        For example, let x and y be points of type II or III.
        If x has center `c_1` and radius `r_1` and y has center
        `c_2` and radius `r_2`, `x < y` if and only if `D(c_1, r_1)`
        is a subset of `D(c_2, r_2)` in `\CC_p`.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point

        OUTPUT:

        - ``True`` -- if this point is greater than ``other`` in the standard partial order
        - ``False`` -- otherwise

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(QQ, 3)
            sage: Q1 = B(5, 3)
            sage: Q2 = B(5, 1)
            sage: Q1.gt(Q2)
            True

        ::

            sage: Q3 = B(1/27)
            sage: Q1.gt(Q3)
            False

        TESTS::

            sage: B = Berkovich_Cp_Projective(QQ, 3)
            sage: Q1 = B(5)
            sage: Q1.gt(Q1)
            False

        ::

            sage: Q2 = B(0, 1)
            sage: Q1.gt(Q2)
            False

        ::

            sage: Q3 = B([0, 3], [5, 1])
            sage: Q2.gt(Q3)
            True

        ::

            sage: Q4 = B((1,0))
            sage: Q4.gt(Q2)
            True

        ::

            sage: Q1.gt(Q4)
            False
        """
        if not isinstance(other, Berkovich_Element_Cp_Projective):
            raise TypeError('other must be a point of a projective Berkovich space, but was %s' % other)
        if self.parent() != other.parent():
            raise ValueError('other must be a point of the same projective Berkovich space')

        if self == other:
            return False
        # infinity is maximal with respect to the standard partial order
        infinity = self.parent()((1, 0))
        if self == infinity:
            return True
        if other == infinity:
            return False

        if self.type_of_point() in [1, 4]:
            return False
        if other.type_of_point() == 4:
            center = other.center()[-1]
            dist = self._custom_abs(self.center()[0] - center[0])
            return dist <= self.radius() and other.radius()[-1] <= self.radius()
        dist = self._custom_abs(self.center()[0] - other.center()[0])
        return dist <= self.radius() and other.radius() <= self.radius()

    def join(self, other, basepoint=Infinity):
        """
        Compute the join of this point and ``other``, with respect to ``basepoint``.

        The join is first point that lies on the intersection
        of the path from this point to ``basepoint`` and the path from ``other`` to
        ``basepoint``.

        INPUT:

        - ``other`` -- a point of the same Berkovich space as this point
        - ``basepoint`` -- (default: ``Infinity``) a point of the same
          Berkovich space as this point, or ``Infinity``

        OUTPUT: a point of the same Berkovich space

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(2, 1)
            sage: Q2 = B(2, 2)
            sage: Q1.join(Q2)
            Type III point centered at (2 + O(3^20) : 1 + O(3^20)) of radius 2.00000000000000

        ::

            sage: Q3 = B(5)
            sage: Q3.join(Q1)
            Type II point centered at (2 + 3 + O(3^20) : 1 + O(3^20)) of radius 3^0

        ::

            sage: Q3.join(Q1, basepoint=Q2)
            Type II point centered at (2 + O(3^20) : 1 + O(3^20)) of radius 3^0

        TESTS::

            sage: Q4 = B(1/3**8 + 2, 1)
            sage: Q2.join(Q4, basepoint=Q1)
            Type III point centered at (2 + O(3^20) : 1 + O(3^20)) of radius 2.00000000000000
            sage: Q5 = B(2, 1/9)
            sage: Q6 = B(1, 1/27)
            sage: Q4.join(Q5, basepoint=Q6)
            Type II point centered at (1 + O(3^20) : 1 + O(3^20)) of radius 3^0
            sage: Q7 = B(1/27, 1/27)
            sage: Q1.join(Q7, Q2)
            Type III point centered at (2 + O(3^20) : 1 + O(3^20)) of radius 2.00000000000000
            sage: Q1.join(Q2, Q7)
            Type III point centered at (2 + O(3^20) : 1 + O(3^20)) of radius 2.00000000000000
            sage: Q8 = B(0, power=1/3)
            sage: Q9 = B(0, power=1/2)
            sage: Q8.join(Q9)
            Type II point centered at (0 : 1 + O(3^20)) of radius 3^1/2

            sage: # needs sage.rings.number_field
            sage: R.<x> = QQ[]
            sage: A.<a> = NumberField(x^3 + 20)
            sage: ideal = A.prime_above(3)
            sage: C = Berkovich_Cp_Projective(A, ideal)
            sage: Q10 = C(a, 1/9)
            sage: Q10.join(Q9)
            Traceback (most recent call last):
            ...
            ValueError: other must be a point of the same projective Berkovich line
            sage: Q11 = C(0, 1/3)
            sage: Q11.join(Q10)
            Type II point centered at (0 : 1) of radius 3^0
        """
        if not isinstance(other, Berkovich_Element_Cp_Projective):
            raise TypeError('other must be a point of a projective Berkovich line, instead was %s' % other)
        if other.parent() != self.parent():
            raise ValueError('other must be a point of the same projective Berkovich line')

        # if either self or other is type IV, we use the last disk in the approximation
        if self.type_of_point() == 4:
            new_center = self.center()[-1]
            new_radius = self.radius()[-1]
            return self.parent()(new_center, new_radius).join(other)
        if other.type_of_point() == 4:
            new_center = other.center()[-1]
            new_radius = other.radius()[-1]
            return self.join(self.parent()(new_center, new_radius))

        # we deal with the point at infinity as a special case
        infty = self.parent()((1, 0))

        if basepoint == Infinity or basepoint == infty:
            if self == infty or other == infty:
                return infty
            dist = self._custom_abs(self.center()[0] - other.center()[0])
            maximum = max(dist, self.radius(), other.radius())
            # optimize for when self or other are type II
            if maximum == self.radius() and self.type_of_point() == 2:
                return self.parent()(self.center(), power=self.power())
            if maximum == other.radius() and other.type_of_point() == 2:
                return self.parent()(self.center(), power=other.power())
            return self.parent()(self.center(), maximum)

        if not isinstance(basepoint, Berkovich_Element_Cp_Projective):
            raise TypeError('basepoint must be a point of a projective Berkovich line, instead was %s' % basepoint)
        if basepoint.parent() != self.parent():
            raise ValueError("basepoint must be a point of the same Berkovich projective line")

        # if the basepoint is type IV, we use the last disk in the approximation
        if basepoint.type_of_point() == 4:
            new_center = other.center()[-1]
            new_radius = other.radius()[-1]
            return self.join(other, self.parent()(new_center, new_radius))

        if self == infty:
            return other.join(basepoint)
        if other == infty:
            return self.join(basepoint)

        b_ge_s = basepoint.gt(self) or basepoint == self
        b_lt_s = basepoint.lt(self)
        b_ge_o = basepoint.gt(other) or basepoint == other
        b_lt_o = basepoint.lt(other)
        s_ge_o = self.gt(other) or self == other
        s_lt_o = self.lt(other)

        # we deal with all the cases where self and other are not comparable first
        if not (s_lt_o or s_ge_o):
            if not (b_ge_o or b_lt_o):
                if not (b_ge_s or b_lt_s):
                    # case where none of the points are comparable
                    dist_b_s = self._custom_abs(self.center()[0] - basepoint.center()[0])
                    dist_b_o = self._custom_abs(other.center()[0] - basepoint.center()[0])
                    return self.parent()(basepoint.center(),
                                         min(max(dist_b_o, other.radius(), basepoint.radius()),
                                             max(dist_b_s, self.radius(), basepoint.radius())))

                # case where self and basepoint are comparable
                if b_ge_s:
                    return basepoint
                return self

            # case where other and basepoint are comparable
            if b_ge_o:
                return basepoint
            return other

        # now the cases where self > other
        if s_ge_o:
            if not (b_ge_s or b_lt_s):
                return self
            if b_ge_s:
                return self
            if b_ge_o:
                return basepoint
            if b_lt_o:
                return other

        # join is symmetric, so we flip self and other so that self > other
        else:
            return other.join(self, basepoint)

    def involution_map(self):
        r"""
        Return the image of this point under the involution map.

        The involution map is the extension of the map ``z |-> 1/z``
        on `P^1(\CC_p)` to Berkovich space.

        If zero is contained in every disk approximating a type IV point,
        then the image under the involution map is not defined. To avoid
        this error, increase precision.

        OUTPUT: a point of the same Berkovich space

        EXAMPLES:

        The involution map is 1/z on type I points::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(1/2)
            sage: Q1.involution_map()
            Type I point centered at (2 + O(3^20) : 1 + O(3^20))

        ::

            sage: Q2 = B(0, 1/3)
            sage: Q2.involution_map()
            Type II point centered at (0 : 1 + O(3^20)) of radius 3^1

        ::

            sage: Q3 = B(1/3, 1/3)
            sage: Q3.involution_map()
            Type II point centered at (3 + O(3^21) : 1 + O(3^20)) of radius 3^-3

        TESTS::

            sage: B = Berkovich_Cp_Projective(3)
            sage: B((1,0)).involution_map()
            Type I point centered at (0 : 1 + O(3^20))

        ::

            sage: B(0).involution_map()
            Type I point centered at (1 + O(3^20) : 0)

        ::

            sage: B(1/81, 1.5).involution_map()
            Type III point centered at (3^4 + O(3^24) : 1 + O(3^20)) of radius 0.000228623685413809

        ::

            sage: B([1, 2], [3, 1]).involution_map()
            Traceback (most recent call last):
            ...
            ValueError: precision of type IV is not high enough to define image

        ::

            sage: B([1/81, 10/81], [10, 9]).involution_map()
            Type IV point of precision 2, approximated by disks centered at
            [(3^4 + O(3^24) : 1 + O(3^20)), (3^4 + 2*3^6 + 2*3^7 + 2*3^10 + 2*3^11 +
            2*3^14 + 2*3^15 + 2*3^18 + 2*3^19 + 2*3^22 + 2*3^23 + O(3^24) : 1 + O(3^20))]
            ... with radii [0.00152415790275873, 0.00137174211248285] ...
        """
        infty = self.parent()((1, 0))
        zero = self.parent()(0)

        if self.type_of_point() == 1:
            if self == infty:
                return zero
            if self == zero:
                return infty
            return self.parent()(1 / self.center()[0])

        if self.type_of_point() in [2, 3]:
            zero_contained_in_self = self.gt(zero)
            if zero_contained_in_self:
                if self.type_of_point() == 2:
                    power = self.power()
                    return self.parent()(ZZ(0), power=-power)
                return self.parent()(ZZ(0), 1 / self.radius())
            return self.parent()(1 / self.center()[0], self.radius() / (self._custom_abs(self.center()[0])**2))

        new_center_lst = []
        new_radius_lst = []
        for i in range(len(self.center())):
            berk_point = self.parent()(self.center()[i], self.radius()[i])
            zero_check = berk_point.gt(zero)
            if zero_check:
                continue
            else:
                new_center = 1 / self.center()[i][0]
                new_radius = self.radius()[i] / (self._custom_abs(self.center()[i][0])**2)
                new_center_lst.append(new_center)
                new_radius_lst.append(new_radius)
        if not new_center_lst:
            raise ValueError('precision of type IV is not high enough to define image')
        return self.parent()(new_center_lst, new_radius_lst)

    def contained_in_interval(self, start, end):
        """
        Check if this point is an element of the interval [``start``, ``end``].

        INPUT:

        - ``start`` -- a point of the same Berkovich space as this point
        - ``end`` -- a point of the same Berkovich space as this point

        OUTPUT:

        - ``True`` if this point is an element of [``start``, ``end``].
        - ``False`` otherwise.

        EXAMPLES::

            sage: B = Berkovich_Cp_Projective(3)
            sage: Q1 = B(2, 1)
            sage: Q2 = B(2, 4)
            sage: Q3 = B(1/3)
            sage: Q2.contained_in_interval(Q1, Q3.join(Q1))
            False

        ::

            sage: Q4 = B(1/81, 1)
            sage: Q2.contained_in_interval(Q1, Q4.join(Q1))
            True

        TESTS::

            sage: B = Berkovich_Cp_Projective(3)
            sage: infty = B((1, 0))
            sage: zero = B(0)
            sage: gauss = B(0, 1)
            sage: infty.contained_in_interval(zero, gauss)
            False

        ::

            sage: Q1 = B(1, 3)
            sage: infty.contained_in_interval(gauss, Q1)
            False

        ::

            sage: zero.contained_in_interval(infty, gauss)
            False

        ::

            sage: gauss.contained_in_interval(zero, infty)
            True

        ::

            sage: Q2 = B(81, 1/3)
            sage: gauss.contained_in_interval(infty, Q2)
            True
        """
        if not isinstance(start, Berkovich_Element_Cp_Projective):
            raise TypeError("start must be a point of Berkovich space")
        if start.parent() != self.parent():
            raise ValueError("start must be a point of the same Berkovich space as this point")
        if not isinstance(end, Berkovich_Element_Cp_Projective):
            raise TypeError("start must be a point of Berkovich space")
        if end.parent() != self.parent():
            raise ValueError("start must be a point of the same Berkovich space as this point")

        # we treat infinity as a special case
        infty = self.parent()((1, 0))
        zero = self.parent()(ZZ(0))
        if self == infty:
            if start == zero or end == zero:
                return end == infty or start == infty
            return (self.involution_map()).contained_in_interval(start.involution_map(),
                                                                 end.involution_map())
        if start == infty or end == infty:
            if self == zero:
                return end == zero or start == zero
            if start == zero or end == zero:
                gauss = self.parent()(ZZ(0), ZZ(1))
                return self.contained_in_interval(start, gauss) or self.contained_in_interval(gauss, end)
            return self.involution_map().contained_in_interval(start.involution_map(),
                                                               end.involution_map())
        join = start.join(end)
        j_ge_s = join.gt(self) or join == self
        s_ge_start = self.gt(start) or self == start
        s_ge_end = self.gt(end) or self == end
        return j_ge_s and (s_ge_end or s_ge_start)



"""
Constructors for certain modular abelian varieties

AUTHORS:

- William Stein (2007-03)
"""
# #########################################################################
#       Copyright (C) 2007 William Stein <wstein@gmail.com>               #
#  Distributed under the terms of the GNU General Public License (GPL)    #
#                  https://www.gnu.org/licenses/                           #
# #########################################################################

import weakref

from sage.rings.integer import Integer

from sage.modular.arithgroup.congroup_gamma0 import Gamma0_constructor as Gamma0
from sage.modular.arithgroup.congroup_generic import CongruenceSubgroupBase
from sage.modular.modsym.space import ModularSymbolsSpace
from .abvar_newform import ModularAbelianVariety_newform
import sage.modular.modform.element
from sage.modular.abvar import abvar

_cache = {}


def _get(key):
    """
    Return the cached abelian variety with given key. This is used
    internally by the abelian varieties constructor.

    INPUT:

    - ``key`` -- hashable

    EXAMPLES::

        sage: sage.modular.abvar.constructor._saved('a', J0(37))
        Abelian variety J0(37) of dimension 2
        sage: sage.modular.abvar.constructor._get('a')
        Abelian variety J0(37) of dimension 2
        sage: sage.modular.abvar.constructor._get('b')
        Traceback (most recent call last):
        ...
        ValueError: element not in cache
    """
    if key in _cache:
        z = _cache[key]()
        if z is not None:
            return z
    raise ValueError("element not in cache")


def _saved(key, J):
    """
    Return the cached abelian variety with given key. This is used
    internally by the abelian varieties constructor.

    INPUT:

    - ``key`` -- hashable

    - ``J`` -- modular abelian variety

    OUTPUT: ``J`` -- returns the modabvar, to make code that uses
    this simpler

    EXAMPLES::

        sage: sage.modular.abvar.constructor._saved('37', J0(37))
        Abelian variety J0(37) of dimension 2
    """
    _cache[key] = weakref.ref(J)
    return J


def J0(N):
    """
    Return the Jacobian `J_0(N)` of the modular curve
    `X_0(N)`.

    EXAMPLES::

        sage: J0(389)
        Abelian variety J0(389) of dimension 32

    The result is cached::

        sage: J0(33) is J0(33)
        True
    """
    key = 'J0(%s)' % N
    try:
        return _get(key)
    except ValueError:
        from sage.modular.arithgroup.congroup_gamma0 import Gamma0_constructor as Gamma0
        J = Gamma0(N).modular_abelian_variety()
        return _saved(key, J)


def J1(N):
    """
    Return the Jacobian `J_1(N)` of the modular curve
    `X_1(N)`.

    EXAMPLES::

        sage: J1(389)
        Abelian variety J1(389) of dimension 6112
    """
    key = 'J1(%s)' % N
    try:
        return _get(key)
    except ValueError:
        from sage.modular.arithgroup.congroup_gamma1 import Gamma1_constructor as Gamma1
        return _saved(key, Gamma1(N).modular_abelian_variety())


def JH(N, H):
    """
    Return the Jacobian `J_H(N)` of the modular curve
    `X_H(N)`.

    EXAMPLES::

        sage: JH(389,[16])
        Abelian variety JH(389,[16]) of dimension 64
    """
    key = 'JH(%s,%s)' % (N, H)
    try:
        return _get(key)
    except ValueError:
        from sage.modular.arithgroup.congroup_gammaH import GammaH_constructor as GammaH
        return _saved(key, GammaH(N, H).modular_abelian_variety())


def AbelianVariety(X):
    """
    Create the abelian variety corresponding to the given defining
    data.

    INPUT:

    - ``X`` -- integer, string, newform, modsym space,
      congruence subgroup or tuple of congruence subgroups

    OUTPUT: a modular abelian variety

    EXAMPLES::

        sage: AbelianVariety(Gamma0(37))
        Abelian variety J0(37) of dimension 2
        sage: AbelianVariety('37a')
        Newform abelian subvariety 37a of dimension 1 of J0(37)
        sage: AbelianVariety(Newform('37a'))
        Newform abelian subvariety 37a of dimension 1 of J0(37)
        sage: AbelianVariety(ModularSymbols(37).cuspidal_submodule())
        Abelian variety J0(37) of dimension 2
        sage: AbelianVariety((Gamma0(37), Gamma0(11)))
        Abelian variety J0(37) x J0(11) of dimension 3
        sage: AbelianVariety(37)
        Abelian variety J0(37) of dimension 2
        sage: AbelianVariety([1,2,3])
        Traceback (most recent call last):
        ...
        TypeError: X must be an integer, string, newform, modsym space, congruence subgroup or tuple of congruence subgroups
    """
    if isinstance(X, (int, Integer)):
        X = Gamma0(X)
    if isinstance(X, CongruenceSubgroupBase):
        X = X.modular_symbols().cuspidal_submodule()
    elif isinstance(X, str):
        from sage.modular.modform.constructor import Newform
        f = Newform(X, names='a')
        return ModularAbelianVariety_newform(f, internal_name=True)
    elif isinstance(X, sage.modular.modform.element.Newform):
        return ModularAbelianVariety_newform(X)

    if isinstance(X, ModularSymbolsSpace):
        return abvar.ModularAbelianVariety_modsym(X)

    if isinstance(X, (tuple, list)) and all(isinstance(G, CongruenceSubgroupBase) for G in X):
        return abvar.ModularAbelianVariety(X)

    raise TypeError("X must be an integer, string, newform, modsym space, congruence subgroup or tuple of congruence subgroups")



"""
Periods of modular forms
"""

# The following idea just occurred to me.
# We could use that $\langle T_n(g), x\rangle = \langle g, T_n(x)\rangle$
# for any Hecke operator $T_n$, so that we only need to compute
# the period integrals $\langle g, x_i\rangle$.  Then we obtain all pairings
# $\langle T_n(g), x_i \rangle = \langle g , T_n(x_i) \rangle$.
# Since the $T_n(g)$ span the simple $\T$-module $S_k(\Gamma;\Q)[I]$,
# this must give all pairings.   However, it requires computing
# only $2d$ pairings instead of $2d^2$ pairings, which is potentially
# a huge savings when $d$ is large.



from __future__ import annotations
from collections.abc import Callable
from typing import Any

class Function:
    def __init__(self, name: str, nargs: int, latex_name: str | None = None, conversions: dict[str, Any] | None = None, evalf_params_first: bool = True, alt_name: str | None = None):
        ...

    def _evalf_try_(self, *args: Any) -> Any:
        ...

    def __hash__(self) -> int:
        ...

    def __repr__(self) -> str:
        ...

    def _latex_(self) -> str:
        ...

    def __richcmp__(self, other: Any, op: int) -> bool:
        ...

    def __call__(self, *args: Any, coerce: bool = True, hold: bool = False) -> Any:
        ...

    def name(self) -> str:
        ...

    def number_of_arguments(self) -> int:
        ...

    def variables(self) -> tuple:
        ...

    def default_variable(self) -> Any:
        ...

    def _is_numerical(self, x: Any) -> bool:
        ...

    def _interface_init_(self, I: Any = None) -> str:
        ...

    def _mathematica_init_(self) -> str:
        ...

    def _sympy_init_(self, I: Any = None) -> str:
        ...

    @property
    def _sympy_(self) -> Callable:
        ...

    def _maxima_init_(self, I: Any = None) -> str:
        ...

    def _fast_callable_(self, etb: Any) -> Any:
        ...

    def _eval_numpy_(self, *args: Any) -> Any:
        ...

    def _eval_mpmath_(self, *args: Any) -> Any:
        ...

class GinacFunction(Function):
    def __init__(self, name: str, nargs: int = 1, latex_name: str | None = None, conversions: dict[str, Any] | None = None, ginac_name: str | None = None, evalf_params_first: bool = True, preserved_arg: int | None = None, alt_name: str | None = None):
        ...

class BuiltinFunction(Function):
    def __init__(self, name: str, nargs: int = 1, latex_name: str | None = None, conversions: dict[str, Any] | None = None, evalf_params_first: bool = True, alt_name: str | None = None, preserved_arg: int | None = None):
        ...

    def _method_arguments(self, arg: Any) -> list:
        ...

    def __call__(self, *args: Any, coerce: bool = True, hold: bool = False, dont_call_method_on_arg: bool = False) -> Any:
        ...

    def _is_registered(self) -> bool:
        ...

    def _evalf_or_eval_(self, *args: Any) -> Any:
        ...

    def __reduce__(self) -> tuple:
        ...

    def __setstate__(self, state: list):
        ...

class SymbolicFunction(Function):
    def __init__(self, name: str, nargs: int = 0, latex_name: str | None = None, conversions: dict[str, Any] | None = None, evalf_params_first: bool = True):
        ...

    def _is_registered(self) -> bool:
        ...

    def _hash_(self) -> int:
        ...

    def __hash__(self) -> int:
        ...

    def __getstate__(self) -> tuple:
        ...

    def __setstate__(self, state: tuple):
        ...

def pickle_wrapper(f: Callable | None) -> bytes | None:
    ...

def unpickle_wrapper(p: bytes | None) -> Callable | None:
    ...



# distutils: libraries = flint
# distutils: depends = flint/fexpr.h

################################################################################
# This file is auto-generated by the script
#   SAGE_ROOT/tools/flint-autogen/flint_autogen.py.
# From the commit 3e2c3a3e091106a25ca9c6fba28e02f2cbcd654a
# Do not modify by hand! Fix and rerun the script instead.
################################################################################

from libc.stdio cimport FILE
from sage.libs.gmp.types cimport *
from sage.libs.mpfr.types cimport *
from sage.libs.flint.types cimport *

cdef extern from "flint_wrap.h":
    void fexpr_init(fexpr_t expr) noexcept
    void fexpr_clear(fexpr_t expr) noexcept
    fexpr_ptr _fexpr_vec_init(slong len) noexcept
    void _fexpr_vec_clear(fexpr_ptr vec, slong len) noexcept
    void fexpr_fit_size(fexpr_t expr, slong size) noexcept
    void fexpr_set(fexpr_t res, const fexpr_t expr) noexcept
    void fexpr_swap(fexpr_t a, fexpr_t b) noexcept
    slong fexpr_depth(const fexpr_t expr) noexcept
    slong fexpr_num_leaves(const fexpr_t expr) noexcept
    slong fexpr_size(const fexpr_t expr) noexcept
    slong fexpr_size_bytes(const fexpr_t expr) noexcept
    slong fexpr_allocated_bytes(const fexpr_t expr) noexcept
    bint fexpr_equal(const fexpr_t a, const fexpr_t b) noexcept
    bint fexpr_equal_si(const fexpr_t expr, slong c) noexcept
    bint fexpr_equal_ui(const fexpr_t expr, ulong c) noexcept
    ulong fexpr_hash(const fexpr_t expr) noexcept
    int fexpr_cmp_fast(const fexpr_t a, const fexpr_t b) noexcept
    bint fexpr_is_integer(const fexpr_t expr) noexcept
    bint fexpr_is_symbol(const fexpr_t expr) noexcept
    bint fexpr_is_string(const fexpr_t expr) noexcept
    bint fexpr_is_atom(const fexpr_t expr) noexcept
    void fexpr_zero(fexpr_t res) noexcept
    bint fexpr_is_zero(const fexpr_t expr) noexcept
    bint fexpr_is_neg_integer(const fexpr_t expr) noexcept
    void fexpr_set_si(fexpr_t res, slong c) noexcept
    void fexpr_set_ui(fexpr_t res, ulong c) noexcept
    void fexpr_set_fmpz(fexpr_t res, const fmpz_t c) noexcept
    int fexpr_get_fmpz(fmpz_t res, const fexpr_t expr) noexcept
    void fexpr_set_symbol_builtin(fexpr_t res, slong id) noexcept
    bint fexpr_is_builtin_symbol(const fexpr_t expr, slong id) noexcept
    bint fexpr_is_any_builtin_symbol(const fexpr_t expr) noexcept
    void fexpr_set_symbol_str(fexpr_t res, const char * s) noexcept
    char * fexpr_get_symbol_str(const fexpr_t expr) noexcept
    void fexpr_set_string(fexpr_t res, const char * s) noexcept
    char * fexpr_get_string(const fexpr_t expr) noexcept
    void fexpr_write(calcium_stream_t stream, const fexpr_t expr) noexcept
    void fexpr_print(const fexpr_t expr) noexcept
    char * fexpr_get_str(const fexpr_t expr) noexcept
    void fexpr_write_latex(calcium_stream_t stream, const fexpr_t expr, ulong flags) noexcept
    void fexpr_print_latex(const fexpr_t expr, ulong flags) noexcept
    char * fexpr_get_str_latex(const fexpr_t expr, ulong flags) noexcept
    slong fexpr_nargs(const fexpr_t expr) noexcept
    void fexpr_func(fexpr_t res, const fexpr_t expr) noexcept
    void fexpr_view_func(fexpr_t view, const fexpr_t expr) noexcept
    void fexpr_arg(fexpr_t res, const fexpr_t expr, slong i) noexcept
    void fexpr_view_arg(fexpr_t view, const fexpr_t expr, slong i) noexcept
    void fexpr_view_next(fexpr_t view) noexcept
    bint fexpr_is_builtin_call(const fexpr_t expr, slong id) noexcept
    bint fexpr_is_any_builtin_call(const fexpr_t expr) noexcept
    void fexpr_call0(fexpr_t res, const fexpr_t f) noexcept
    void fexpr_call1(fexpr_t res, const fexpr_t f, const fexpr_t x1) noexcept
    void fexpr_call2(fexpr_t res, const fexpr_t f, const fexpr_t x1, const fexpr_t x2) noexcept
    void fexpr_call3(fexpr_t res, const fexpr_t f, const fexpr_t x1, const fexpr_t x2, const fexpr_t x3) noexcept
    void fexpr_call4(fexpr_t res, const fexpr_t f, const fexpr_t x1, const fexpr_t x2, const fexpr_t x3, const fexpr_t x4) noexcept
    void fexpr_call_vec(fexpr_t res, const fexpr_t f, fexpr_srcptr args, slong len) noexcept
    void fexpr_call_builtin1(fexpr_t res, slong f, const fexpr_t x1) noexcept
    void fexpr_call_builtin2(fexpr_t res, slong f, const fexpr_t x1, const fexpr_t x2) noexcept
    bint fexpr_contains(const fexpr_t expr, const fexpr_t x) noexcept
    int fexpr_replace(fexpr_t res, const fexpr_t expr, const fexpr_t x, const fexpr_t y) noexcept
    int fexpr_replace2(fexpr_t res, const fexpr_t expr, const fexpr_t x1, const fexpr_t y1, const fexpr_t x2, const fexpr_t y2) noexcept
    int fexpr_replace_vec(fexpr_t res, const fexpr_t expr, const fexpr_vec_t xs, const fexpr_vec_t ys) noexcept
    void fexpr_set_fmpq(fexpr_t res, const fmpq_t x) noexcept
    void fexpr_set_arf(fexpr_t res, const arf_t x) noexcept
    void fexpr_set_d(fexpr_t res, double x) noexcept
    void fexpr_set_re_im_d(fexpr_t res, double x, double y) noexcept
    void fexpr_neg(fexpr_t res, const fexpr_t a) noexcept
    void fexpr_add(fexpr_t res, const fexpr_t a, const fexpr_t b) noexcept
    void fexpr_sub(fexpr_t res, const fexpr_t a, const fexpr_t b) noexcept
    void fexpr_mul(fexpr_t res, const fexpr_t a, const fexpr_t b) noexcept
    void fexpr_div(fexpr_t res, const fexpr_t a, const fexpr_t b) noexcept
    void fexpr_pow(fexpr_t res, const fexpr_t a, const fexpr_t b) noexcept
    bint fexpr_is_arithmetic_operation(const fexpr_t expr) noexcept
    void fexpr_arithmetic_nodes(fexpr_vec_t nodes, const fexpr_t expr) noexcept
    int fexpr_get_fmpz_mpoly_q(fmpz_mpoly_q_t res, const fexpr_t expr, const fexpr_vec_t vars, const fmpz_mpoly_ctx_t ctx) noexcept
    void fexpr_set_fmpz_mpoly(fexpr_t res, const fmpz_mpoly_t poly, const fexpr_vec_t vars, const fmpz_mpoly_ctx_t ctx) noexcept
    void fexpr_set_fmpz_mpoly_q(fexpr_t res, const fmpz_mpoly_q_t frac, const fexpr_vec_t vars, const fmpz_mpoly_ctx_t ctx) noexcept
    int fexpr_expanded_normal_form(fexpr_t res, const fexpr_t expr, ulong flags) noexcept
    void fexpr_vec_init(fexpr_vec_t vec, slong len) noexcept
    void fexpr_vec_clear(fexpr_vec_t vec) noexcept
    void fexpr_vec_print(const fexpr_vec_t vec) noexcept
    void fexpr_vec_swap(fexpr_vec_t x, fexpr_vec_t y) noexcept
    void fexpr_vec_fit_length(fexpr_vec_t vec, slong len) noexcept
    void fexpr_vec_set(fexpr_vec_t dest, const fexpr_vec_t src) noexcept
    void fexpr_vec_append(fexpr_vec_t vec, const fexpr_t expr) noexcept
    slong fexpr_vec_insert_unique(fexpr_vec_t vec, const fexpr_t expr) noexcept
    void fexpr_vec_set_length(fexpr_vec_t vec, slong len) noexcept
    void _fexpr_vec_sort_fast(fexpr_ptr vec, slong len) noexcept



r"""
This module contains Cython code for a backtracking algorithm to solve Sudoku puzzles.

Once Cython implements closures and the ``yield`` keyword is possible, this can be moved into the ``sage.games.sudoku`` module, as part of the ``Sudoku.backtrack`` method, and this module can be banned.
"""


def backtrack_all(n, puzzle):
    r"""
    A routine to compute all the solutions to a Sudoku puzzle.

    INPUT:

        - ``n`` -- the size of the puzzle, where the array is an `n^2\times n^2` grid

        - ``puzzle`` -- list of the entries of the puzzle (1-based), in row-major order

    OUTPUT:

        A list of solutions, where each solution is a (1-based) list similar to ``puzzle``.

    TESTS:

    This is just a cursory test here, since eventually this code will move.
    See the `backtrack` method of the `Sudoku` class in the
    `sage.games.sudoku` module for more enlightening examples. ::

        sage: from sage.games.sudoku_backtrack import backtrack_all
        sage: c = [0, 0, 0, 0, 1, 0, 9, 0, 0, 8, 0, 0, 4, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 7, 0, 0, 3, 0, 0, 0,0, 0, 0, 0, 0, 0, 0, 2, 0, 4, 0, 0, 0, 0, 0, 0, 0, 5, 8, 0, 6, 0, 0, 0, 0, 1, 3, 0, 7, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 8, 0, 0, 0, 0, 0]
        sage: print(backtrack_all(3, c))
        [[6, 5, 4, 3, 1, 2, 9, 8, 7, 8, 3, 1, 4, 7, 9, 5, 2, 6, 2, 9, 7, 6, 8, 5, 4, 1, 3, 4, 7, 2, 5, 3, 8, 6, 9, 1, 3, 8, 5, 1, 9, 6, 2, 7, 4, 9, 1, 6, 7, 2, 4, 3, 5, 8, 5, 6, 8, 9, 4, 7, 1, 3, 2, 7, 4, 3, 2, 5, 1, 8, 6, 9, 1, 2, 9, 8, 6, 3, 7, 4, 5]]

    ALGORITHM:

    We traverse a search tree depth-first.  Each level of the tree corresponds to a location in the grid, listed in row-major order.  At each location we maintain a list of the symbols which may be used in that location as follows.

    A location has "peers", which are the locations in the same row, column or box (sub-grid).  As symbols are chosen (or fixed initially) at a location, they become ineligible for use at a peer.  We track this in the ``available`` array where at each location each symbol has a count of how many times it has been made ineligible.  As this counter transitions between 0 and 1, the number of eligible symbols at a location is tracked in the ``card`` array.  When the number of eligible symbols at any location becomes 1, then we know that *must* be the symbol employed in that location.  This then allows us to further update the eligible symbols at the peers of that location.  When the number of the eligible symbols at any location becomes 0, then we know that we can prune the search tree.

    So at each new level of the search tree, we propagate as many fixed symbols as we can, placing them into a two-ended queue (``fixed`` and ``fixed_symbol``) that we process until it is empty or we need to prune.  All this recording of ineligible symbols and numbers of eligible symbols has to be unwound as we backup the tree, though.

    The notion of propagating singleton cells forward comes from an essay by Peter Norvig [sudoku:norvig]_.
    """
    cdef:
        # Arrays sizes are n^4, and n^2, with 3n^2-2n-1 for second slot of peers, n = 4
        int i, j, count, level, apeer
        int nsquare, npeers, nboxes
        int grid_row, grid_col, grid_corner
        int peers[256][39]
        int box[256]
        int available[256][16]
        int card[256]
        int hint, symbol, abox
        int feasible
        int nfixed[256]
        int fixed[256][256]
        int fixed_symbol[256][256]

        int process, asymbol, alevel, process_level, process_symbol

    # sanity check on size (types)
    # n is "base" dimension
    # nsquare is size of a grid
    # nboxes is total number of entries
    nsquare = n*n
    nboxes = nsquare * nsquare
    npeers = 3*n*n-2*n-1  # 2(n^2-1)+n^2-2n+1

    # "Peers" of a box are boxes in the same column, row or grid
    # Like the conflict graph when expressed as a graph coloring problem
    for level in range(nboxes):
        # location as row and column in square
        # grids are numbered similarly, in row-major order
        row = level // nsquare
        col = level % nsquare
        grid_corner = (row - (row % n))*nsquare + (col - (col % n))
        grid_row = row // n
        grid_col = col // n
        count = -1
        # Peers' levels in same grid, but not the box itself
        for i in range(n):
            for j in range(n):
                grid_level = grid_corner + i*nsquare + j
                if grid_level != level:
                    count += 1
                    peers[level][count] = grid_level
        # Peers' levels in the same row, but not in grid
        for i in range(nsquare):
            if (i // 3 != grid_col):
                count += 1
                peers[level][count] = row*nsquare + i
        # Peers' levels in same column, but not in grid
        for i in range(nsquare):
            if (i // 3 != grid_row):
                count += 1
                peers[level][count] = col + i*nsquare

    # Initialize data structures
    # Make every symbol available initially for a box
    # And make set cardinality the size of symbol set
    for level in range(nboxes):
        box[level] = -1
        card[level] = nsquare
        for j in range(nsquare):
            available[level][j] = 0

    # For nonzero entries of input puzzle
    # (1) Convert to zero-based indexing
    # (2) Make a set of size 1 available initially
    for level in range(nboxes):
        # location as row and column in square
        # grids are numbered similarly, in row-major order
        hint = puzzle[level] - 1
        if hint != -1:
            # Limit symbol set at the hint's location to a singleton
            for j in range(nsquare):
                available[level][j] = 1
            available[level][hint] = 0
            card[level] = 1
            #  Remove hint from all peers' available symbols
            #  Track cardinality as sets adjust
            for i in range(npeers):
                apeer = peers[level][i]
                available[apeer][hint] += 1
                if available[apeer][hint] == 1:
                    card[apeer] -= 1

    # Start backtracking
    solutions = []
    level = 0
    box[level] = -1
    while (level > -1):
        symbol = box[level]
        if (symbol != -1):
            # restore symbols to peers
            for i in range(nfixed[level]):
                alevel = fixed[level][i]
                asymbol = fixed_symbol[level][i]
                for j in range(npeers):
                    abox = peers[alevel][j]
                    available[abox][asymbol] -= 1
                    if available[abox][asymbol] == 0:
                        card[abox] += 1
        # move sideways in search tree to next available symbol
        symbol += 1
        while (symbol < nsquare) and (available[level][symbol] != 0):
            symbol += 1
        if symbol == nsquare:
            # fell off the end sideways, backup
            level = level - 1
        else:
            box[level] = symbol
            # Remove elements of sets, adjust cardinalities
            # Descend in search tree if no empty sets created
            # Can't break early at an empty set
            #   or we will confuse restore that happens immediately
            feasible = True
            fixed[level][0] = level
            fixed_symbol[level][0] = symbol
            count = 0
            process = -1
            while (process < count) and feasible:
                process += 1
                process_level = fixed[level][process]
                process_symbol = fixed_symbol[level][process]
                for i in range(npeers):
                    abox = peers[process_level][i]
                    available[abox][process_symbol] += 1
                    if available[abox][process_symbol] == 1:
                        card[abox] -= 1
                        if card[abox] == 0:
                            feasible = False
                        if card[abox] == 1:
                            count += 1
                            fixed[level][count] = abox
                            asymbol = 0
                            while (available[abox][asymbol] != 0):
                                asymbol += 1
                            fixed_symbol[level][count] = asymbol
            nfixed[level] = process+1
            if feasible:
                if level == nboxes - 1:
                    # Have a solution to save, stay at this bottom-most level
                    # Once Cython implements closures, a yield can go here
                    solutions.append([box[i]+1 for i in range(nboxes)])
                else:
                    level = level + 1
                    box[level] = -1
    return solutions
